// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <fstream>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <opencv/cv.h>
#include <opencv/cvaux.h>
#include <opencv/highgui.h> 

// Local includes
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "homotopy_NN_Astar.h"
#include "D3N2_MovingObstacleIn2D.h"

// ==================================================================
// OBSTACLE DEFINITIONS
// ------------------------------------------------------------------

// DISK:

Obstacle_disk::Obstacle_disk(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent)
{
	shape = "disk";
	parentEnv = parent;
	envFileNode["c"][0] >> cx;
	envFileNode["c"][1] >> cy;
	envFileNode["rx"] >> rx;
	envFileNode["ry"] >> ry;
	getHomotopySkeleton(envFileNode.FindValue("homotopySkeleton"), cx, cy);
}

bool Obstacle_disk::isPointInside(double x, double y, double tt)
{
	arma::vec cUntransformed = _vArma(2, cx.eval(tt), cy.eval(tt));
	double Rx = rx.eval(tt) + padding;
	double Ry = ry.eval(tt) + padding;
	arma::mat transMat = GetTransMatrix(tt);
	arma::vec transVec = GetTransVector(tt);
	
	arma::vec pta = arma::inv(transMat) * ( _vArma(2,x,y) - cUntransformed - transVec ); // Inverse tranaformed coordinate
	return ( pta[0]*pta[0]/(Rx*Rx) + pta[1]*pta[1]/(Ry*Ry) <= 1.0 );
}

void Obstacle_disk::drawInImage_cv(IplImage* ipl_image_p, double tt)
{
	arma::vec cUntransformed = _vArma(2, cx.eval(tt), cy.eval(tt));
	double Rx = rx.eval(tt);
	double Ry = ry.eval(tt);
	arma::mat transMat = GetTransMatrix(tt);
	arma::vec transVec = GetTransVector(tt);
	arma::vec pta;
	std::vector<int> pt;
	CvPoint ptcv;
	CvScalar colr = OBSTACLE_COLOR_CV;
	
	int polyVertCount = 100, vertCount = 0;
	CvPoint* pts = new CvPoint[polyVertCount];
	double xd, yd, angstep=(2*PI_CONST/(double)polyVertCount);
	for (double th=-PI_CONST; th<PI_CONST && vertCount<polyVertCount; th+=angstep)
	{
		xd = Rx * cos(th);
		yd = Ry * sin(th);
		pta = transMat*(_vArma(2,xd,yd)) + cUntransformed + transVec;
		pt = parentEnv->plotTransform.cont2disc( Arma2Std_vec(pta) );
		pts[vertCount] = cvPoint(pt[0], pt[1]);
		vertCount++;
	}
	cvFillConvexPoly (ipl_image_p, pts, polyVertCount, colr);
	delete[] pts;
	
	arma::vec skeletonPos = getSkeletonPos(tt);
	std::vector<int> APoint = parentEnv->plotTransform.cont2disc( Arma2Std_vec(skeletonPos) );
	cvRectangle(ipl_image_p, cvPoint(APoint[0]-1, APoint[1]-1), cvPoint(APoint[0]+1, APoint[1]+1), SKELETON_COLOR_CV);
}

// ---------------------
// RECTANGLE:

Obstacle_rectangle::Obstacle_rectangle(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent)
{
	shape = "rectangle";
	parentEnv = parent;
	envFileNode["c"][0] >> cx;
	envFileNode["c"][1] >> cy;
	envFileNode["lx"] >> lx;
	envFileNode["ly"] >> ly;
	getHomotopySkeleton(envFileNode.FindValue("homotopySkeleton"), cx, cy);
}

bool Obstacle_rectangle::isPointInside(double x, double y, double tt)
{
	arma::vec cUntransformed = _vArma(2, cx.eval(tt), cy.eval(tt));
	double Rx = lx.eval(tt)/2.0 + padding;
	double Ry = ly.eval(tt)/2.0 + padding;
	arma::mat transMat = GetTransMatrix(tt);
	arma::vec transVec = GetTransVector(tt);
	
	arma::vec pta = arma::inv(transMat) * ( _vArma(2,x,y) - cUntransformed - transVec ); // Inverse tranaformed coordinate
	return ( fabs(pta[0])<=Rx && fabs(pta[1])<=Ry );
}

void Obstacle_rectangle::drawInImage_cv(IplImage* ipl_image_p, double tt)
{
	arma::vec cUntransformed = _vArma(2, cx.eval(tt), cy.eval(tt));
	double Rx = lx.eval(tt)/2.0;
	double Ry = ly.eval(tt)/2.0;
	arma::mat transMat = GetTransMatrix(tt);
	arma::vec transVec = GetTransVector(tt);
	arma::vec pta;
	std::vector<int> pt;
	CvPoint ptcv;
	CvScalar colr = OBSTACLE_COLOR_CV;
	
	int vertCount = 0;
	double sgnx;
	CvPoint* pts = new CvPoint[4];
	for (double xd=-Rx; xd<=Rx; xd+=2*Rx)
	{
		sgnx = (double)(_sign(xd));
		for (double yd=-Ry*sgnx; yd*sgnx<=Ry; yd+=2*Ry*sgnx)
		{
			pta = transMat*(_vArma(2,xd,yd)) + cUntransformed + transVec;
			pt = parentEnv->plotTransform.cont2disc( Arma2Std_vec(pta) );
			pts[vertCount] = cvPoint(pt[0], pt[1]);
			vertCount++;
		}
	}
	cvFillConvexPoly (ipl_image_p, pts, 4, colr);
	delete[] pts;
	
	arma::vec skeletonPos = getSkeletonPos(tt);
	std::vector<int> APoint = parentEnv->plotTransform.cont2disc( Arma2Std_vec(skeletonPos) );
	cvRectangle(ipl_image_p, cvPoint(APoint[0]-1, APoint[1]-1), cvPoint(APoint[0]+1, APoint[1]+1), SKELETON_COLOR_CV);
}

// ---------------------
// STATIC_MAP: // generaly "homotopySkeleton" will be "none".

Obstacle_staticmap::Obstacle_staticmap(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent)
{
	shape = "staticmap";
	parentEnv = parent;
	envFileNode["image"] >> filename;
	// load image
	printf("Loading map: %s\n", filename.c_str());
	image_original = cvLoadImage( filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
	// Resize for plotting
	image_plotting = cvCreateImage( cvSize(parentEnv->plotTransform.disc_size[0], parentEnv->plotTransform.disc_size[1]), 
										image_original->depth, image_original->nChannels );
	cvResize( image_original, image_plotting, CV_INTER_AREA );
	// Resize for planning
	image_planning = cvCreateImage( cvSize(parentEnv->planTransform.disc_size[0], parentEnv->planTransform.disc_size[1]), 
										image_original->depth, image_original->nChannels );
	cvResize( image_original, image_planning, CV_INTER_AREA );
	// No homotopy skeleton
}

bool Obstacle_staticmap::isPointInside(double x, double y, double tt)
{
	// This is used only during planning
	std::vector<int> discCoords = parentEnv->planTransform.cont2disc( _vStd_d(2, x, y) );
	if ( discCoords[0]>=parentEnv->planTransform.disc_size[0] || discCoords[1]>=parentEnv->planTransform.disc_size[1] )
		return false;
	else
		return ( ((unsigned char*)(image_planning->imageData + 
					image_planning->widthStep * (parentEnv->planTransform.disc_size[1]-discCoords[1]-1)))
						[discCoords[0]] < 125 );
}

void Obstacle_staticmap::drawInImage_cv(IplImage* ipl_image_p, double tt)
{
	// Used only for plotting
	for (int xi=0; xi<parentEnv->plotTransform.disc_size[0]; xi++)
		for (int yi=0; yi<parentEnv->plotTransform.disc_size[1]; yi++)
		{
			if ( ((unsigned char*)(image_plotting->imageData + 
						image_plotting->widthStep*(parentEnv->plotTransform.disc_size[1]-yi-1)))
							[xi] < 125 )
				cvRectangle(ipl_image_p, cvPoint(xi, yi), cvPoint(xi, yi), OBSTACLE_COLOR_CV);
		}
}


// ---------------------
// POLYGON:

// ------------------------------------------------------------------
// ==================================================================

D3N2_MovingObstacleIn2D_Env::D3N2_MovingObstacleIn2D_Env()
{
	Active_Env_p = this;
}

void D3N2_MovingObstacleIn2D_Env::parseObstacle(const YAML::Node& nObs)
{
	GenericObstacle* thisObstacle = NULL;
	
	if (nObs.FindValue("disk"))
	{
		Obstacle_disk* thisDisk = new Obstacle_disk(nObs["disk"], this);
		thisObstacle = (GenericObstacle*)thisDisk;
	}
	else if (nObs.FindValue("rectangle"))
	{
		Obstacle_rectangle* thisRectangle = new Obstacle_rectangle(nObs["rectangle"], this);
		thisObstacle = (GenericObstacle*)thisRectangle;
	}
	else if (nObs.FindValue("staticmap"))
	{
		Obstacle_staticmap* thisStaticmap = new Obstacle_staticmap(nObs["staticmap"], this);
		thisObstacle = (GenericObstacle*)thisStaticmap;
	}
	// else if
	
	// --------------------------------------------------
	if (thisObstacle)
	{
		//thisObstacle->parentEnv = this;
		thisObstacle->getTransformation( nObs[thisObstacle->shape].FindValue("transform") );
		
		if (nObs[thisObstacle->shape].FindValue("simplexSampleRate"))
			nObs[thisObstacle->shape]["simplexSampleRate"] >> thisObstacle->simplexSampleRate;
		else
		{
			char tmpstr[64];
			sprintf(tmpstr, "%f", cellsize_planner[2]);
			thisObstacle->simplexSampleRate.init(tmpstr);
		}
		
		if (nObs[thisObstacle->shape].FindValue("padding"))
			nObs[thisObstacle->shape]["padding"] >> thisObstacle->padding;
		else
			thisObstacle->padding = 0.0;
		
		obstacles.push_back( thisObstacle );
	}
}

// ******************************************************************
// ==================================================================

muExpression::muExpression(std::string expr)
{
	p = new mu::Parser;
	t = new double;
	p->DefineVar("t", t);
	p->SetExpr(expr);
	s = expr;
}

void muExpression::init(std::string expr)
{
	p = new mu::Parser;
	t = new double;
	p->DefineVar("t", t);
	p->SetExpr(expr);
	s = expr;
}

double muExpression::eval(double tt)
{
	*t = tt;
	return ( p->Eval() );
}

void operator >> (const YAML::Node& node, muExpression& ex)
{
	std::string expr;
	node >> expr;
	ex.init(expr);
}

//-------------------------------------------

Traj_XYT::Traj_XYT(int nn)
{
	n = nn;
	x = new double[n];
	y = new double[n];
	t = new double[n];
	interp_initiated = false;
}

void Traj_XYT::_free(void)
{
	delete[] x;
	delete[] y;
	delete[] t;
}

std::vector< double > Traj_XYT::getCoords(double tt)
{
	if (!interp_initiated)
	{
		workspace_x = gsl_interp_alloc(gsl_interp_linear, n);
		workspace_y = gsl_interp_alloc(gsl_interp_linear, n);
		gsl_interp_init(workspace_x, t, x, n);
		gsl_interp_init(workspace_y, t, y, n);
		interp_initiated = true;
	}
	
	std::vector< double > ret(2);
	ret[0] = gsl_interp_eval(workspace_x, t, x, tt, NULL);
	ret[1] = gsl_interp_eval(workspace_y, t, y, tt, NULL);
	return (ret);
}

//-------------------------------------------

void GenericObstacle::getTransformation(const YAML::Node& envFileNode)
{
	if (envFileNode)
	{
		std::vector< std::vector< muExpression > > theMat(2);
		std::vector< muExpression > theVec(2);
		muExpression tmpexpr;
	
		envFileNode["matrix"][0][0] >> tmpexpr;
		theMat[0].push_back(tmpexpr);
		envFileNode["matrix"][0][1] >> tmpexpr;
		theMat[0].push_back(tmpexpr);
		envFileNode["matrix"][1][0] >> tmpexpr;
		theMat[1].push_back(tmpexpr);
		envFileNode["matrix"][1][1] >> tmpexpr;
		theMat[1].push_back(tmpexpr);
		transMat = theMat;
	
		if (envFileNode["vector"])
		{
			envFileNode["vector"][0] >> tmpexpr;
			theVec[0] = tmpexpr;
			envFileNode["vector"][1] >> tmpexpr;
			theVec[1] = tmpexpr;
		}
		else
		{
			theVec[0].init("0.0");
			theVec[1].init("0.0");
		}
		transVec = theVec;
	}
	else
	{
		
	}
}

void GenericObstacle::getHomotopySkeleton(const YAML::Node& envFileNode, muExpression default_x, muExpression default_y)
{
	muExpression tmpexpr;
	if (envFileNode)
	{
		envFileNode[0] >> tmpexpr;
		if (!tmpexpr.s.compare("none"))
			homotopySkeleton.push_back(tmpexpr);
		else if (!tmpexpr.s.compare("auto"))
		{
			homotopySkeleton.push_back(default_x);
			homotopySkeleton.push_back(default_y);
		}
		else
		{
			homotopySkeleton.push_back(tmpexpr);
			envFileNode[1] >> tmpexpr;
			homotopySkeleton.push_back(tmpexpr);
		}
	}
	else
	{
	
	}
}

// ---------------------------------------------

arma::mat GenericObstacle::GetTransMatrix(double t)
{
	arma::mat thisTransMat(2,2);
	for (int a=0; a<=1; a++)
		for (int b=0; b<=1; b++)
			thisTransMat(a, b) = transMat[a][b].eval(t);
	return (thisTransMat);
}

arma::vec GenericObstacle::GetTransVector(double t)
{
	arma::vec thisTransVec(2);
	for (int a=0; a<=1; a++)
		thisTransVec(a) =  transVec[a].eval(t);
	return (thisTransVec);
}

arma::vec GenericObstacle::getSkeletonPos(double t)
{
	arma::mat thisTransMat = GetTransMatrix(t);
	arma::vec thisTransVec = GetTransVector(t);
	arma::vec thisUntransformedPos = _vArma( 2, homotopySkeleton[0].eval(t), homotopySkeleton[1].eval(t) );
	
	return ( (arma::vec) ( thisUntransformedPos + thisTransVec ) );
}


D3N2_MovingObstacleIn2D_Env::~D3N2_MovingObstacleIn2D_Env()
{
	for (int a=0; a<obstacles.size(); a++)
		delete obstacles[a];
		
	for (int a=0; a<trajectories.size(); a++)
		trajectories[a]._free();
}

// ==================================================================

void D3N2_MovingObstacleIn2D_Env::parseEnvFile(const char* fName, bool vis)
{
	//std::ifstream fin(fName);
	//YAML::Parser parser(fin);
	YAML::Node doc = YAML::LoadFile(fName);
	//parser.GetNextDocument(doc);
	
	// --------------------------------------------------
	doc["problem"]["description"] >> problem_description;
	
	if (!problem_description.compare("Check known trajectory"))
	{
		Traj_XYT knownTraj ( doc["problem"]["trajectory"].size() );
		for (int a=0; a<knownTraj.n; a++)
		{
			knownTraj.x[a] = doc["problem"]["trajectory"][a][0].as<double>();
			knownTraj.y[a] = doc["problem"]["trajectory"][a][1].as<double>();
			knownTraj.t[a] = doc["problem"]["trajectory"][a][2].as<double>();
		}
		trajectories.push_back(knownTraj);
	}
	
	else if (!problem_description.compare("Explore homotopy classes"))
	{
		startcoord.resize(2);
		stopcoord.resize(2);
		doc["problem"]["maxcount"] >> exploreMaxCount;
		doc["problem"]["startcoord"][0] >> startcoord[0];
		doc["problem"]["startcoord"][1] >> startcoord[1];
		doc["problem"]["stopcoord"][0] >> stopcoord[0];
		doc["problem"]["stopcoord"][1] >> stopcoord[1];
		if (doc["problem"]["onlyjordan"])
			doc["problem"]["onlyjordan"] >> onlyJordan;
		else
			onlyJordan = false;
		//printf("onlyjordan = %d\n", onlyJordan);
	}
	
	// For visualization of output file:
	if (doc["outputTrajectories"])
	{
		for (int a=0; a<doc["outputTrajectories"].size(); a++)
		{
			Traj_XYT outTraj ( doc["outputTrajectories"][a]["trajectory"].size() );
			for (int b=0; b<outTraj.n; b++)
			{
				outTraj.x[b] = doc["outputTrajectories"][a]["trajectory"][b][0].as<double>();
				outTraj.y[b] = doc["outputTrajectories"][a]["trajectory"][b][1].as<double>();
				outTraj.t[b] = doc["outputTrajectories"][a]["trajectory"][b][2].as<double>();
			}
			trajectories.push_back(outTraj);
		}
	}
	// --------------------------------------------------
	
	doc["timerange"][0] >> tStart;
	doc["timerange"][1] >> tEnd;
	
	if (doc["plotoptions"])
	{
		std::vector<double> mins(2), maxs(2);
		double cs;
		doc["plotoptions"]["windowsize"][0] >> mins[0];
		doc["plotoptions"]["windowsize"][1] >> mins[1];
		doc["plotoptions"]["windowsize"][2] >> maxs[0];
		doc["plotoptions"]["windowsize"][3] >> maxs[1];
		doc["plotoptions"]["cellsize"] >> cs;
		plotTransform.init(mins, maxs, cs);
		doc["plotoptions"]["timestep"] >> plotTimestep;
	}
	
	if (doc["planoptions"])
	{
		mins_planner.resize(3); maxs_planner.resize(3); cellsize_planner.resize(3);
		doc["planoptions"]["windowsizeXY"][0] >> mins_planner[0];
		doc["planoptions"]["windowsizeXY"][1] >> mins_planner[1];
		doc["planoptions"]["windowsizeXY"][2] >> maxs_planner[0];
		doc["planoptions"]["windowsizeXY"][3] >> maxs_planner[1];
		mins_planner[2] = tStart;
		maxs_planner[2] = tEnd;
		doc["planoptions"]["cellsizeXY"] >> cellsize_planner[0];
		doc["planoptions"]["cellsizeXY"] >> cellsize_planner[1];
		doc["planoptions"]["cellsizeT"] >> cellsize_planner[2];
		planTransform.init(mins_planner, maxs_planner, cellsize_planner);
	}
	
	if (doc["loopClosureParams"])
	{
		doc["loopClosureParams"]["negInfTime"] >> loop_neginf_t;
		doc["loopClosureParams"]["posInfTime"] >> loop_posinf_t;
		doc["loopClosureParams"]["conjugateCenterCoord"][0] >> conj_x;
		doc["loopClosureParams"]["conjugateCenterCoord"][1] >> conj_y;
		doc["loopClosureParams"]["conjugateCenterCoord"][2] >> conj_t;
	}
	
	// Parsing obstacles
	for (int a=0; a<doc["obstacles"].size(); a++)
		parseObstacle(doc["obstacles"][a]);
	
}

// ---------------------------------------------

bool isCoordFree(arma::vec coord)
{
	if ( coord[0]<Active_Env_p->mins_planner[0] || coord[1]<Active_Env_p->mins_planner[1] || coord[2]<Active_Env_p->mins_planner[2] || 
			coord[0]>Active_Env_p->maxs_planner[0] || coord[1]>Active_Env_p->maxs_planner[1] || coord[2]>Active_Env_p->maxs_planner[2]  )
		return false;

	for (int a=0; a<Active_Env_p->obstacles.size(); a++)
		if ( Active_Env_p->obstacles[a]->isPointInside(coord[0], coord[1], coord[2]) )
			return false;
			
	return true;
}

// ---------------------------------------------

arma::mat metricTensor_f(arma::vec coord)
{
	double dx = coord[0] - Active_Env_p->stopcoord[0];
	double dy = coord[1] - Active_Env_p->stopcoord[1];
	if ( dx*dx + dy*dy < Active_Env_p->cellsize_planner[0] * Active_Env_p->cellsize_planner[1] ) // close to goal
		return ( arma::diagmat( _vArma(3, 1.0, 1.0, 1e-8) ) );
	else
		return ( arma::diagmat( _vArma(3, 1.0, 1.0, 1e-4) ) );
}

// ---------------------------------------------

double heuristics_f(arma::vec c1, arma::vec c2)
{
	// 8-connected heuristics
	double dx = fabs(c1[0] - c2[0]);
	double dy = fabs(c1[1] - c2[1]);
	return ( fabs(dx-dy) + SQRT2_CONST*__min(dx,dy) );
}

