#ifndef D3N2_MOVINGOBSTACLESIN2D_H
#define D3N2_MOVINGOBSTACLESIN2D_H

// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <fstream>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <opencv/cv.h>
#include <opencv/cvaux.h>

// Local includes
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "homotopy_NN_Astar.h"

// Definations
#define ROBOT_COLOR_CV CV_RGB(200.0,0.0,0.0)
#define OBSTACLE_COLOR_CV CV_RGB(0.0,80.0,40.0)
#define SKELETON_COLOR_CV CV_RGB(150.0,150.0,0.0)

class muExpression
{
public:
	std::string s;
	mu::Parser* p;
	double* t;
	muExpression() { };
	muExpression(std::string expr);
	void init(std::string expr);
	double eval(double tt);
	// TODO: Add copy constructor, = operator and destructor. Without those there is a memory leak every time 'init' is called.
};

void operator >> (const YAML::Node& node, muExpression& ex);

// ==============================================

class D3N2_MovingObstacleIn2D_Env;

class GenericObstacle
{
public:
	arma::vec getSkeletonPos(double t);
	virtual bool isPointInside(double x, double y, double t) = 0;
	virtual void drawInImage_cv(IplImage* ipl_image_p, double tt) = 0;
	void getTransformation(const YAML::Node& envFileNode);
	void getHomotopySkeleton(const YAML::Node& envFileNode, muExpression default_x, muExpression default_y);
	arma::mat GetTransMatrix(double t);
	arma::vec GetTransVector(double t);
	D3N2_MovingObstacleIn2D_Env* parentEnv;
	// -----------------------
	std::vector< std::vector< muExpression > > transMat;
	std::vector< muExpression > transVec;
	std::vector< muExpression > homotopySkeleton;
	double padding; // to be used only in isPointInside
	muExpression simplexSampleRate;
	// -----------------------
	std::string shape;
};

// ----------------------------

class Obstacle_disk: public GenericObstacle
{
public:
	muExpression cx, cy, rx, ry;
	bool isPointInside(double x, double y, double tt);
	void drawInImage_cv(IplImage* ipl_image_p, double tt);
	Obstacle_disk(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent);
};

class Obstacle_rectangle: public GenericObstacle
{
public:
	muExpression cx, cy, lx, ly;
	bool isPointInside(double x, double y, double tt);
	void drawInImage_cv(IplImage* ipl_image_p, double tt);
	Obstacle_rectangle(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent);
};

class Obstacle_staticmap: public GenericObstacle
{
public:
	std::string filename;
	IplImage* image_original;
	IplImage* image_plotting;
	IplImage* image_planning;
	bool isPointInside(double x, double y, double tt);
	void drawInImage_cv(IplImage* ipl_image_p, double tt);
	Obstacle_staticmap(const YAML::Node& envFileNode, D3N2_MovingObstacleIn2D_Env* parent);
	// TODO: Add a _free() function and call it for destruction.
};

// ===============================================

class Traj_XYT
{
public:
	int n;
	double* x;
	double* y;
	double* t;
	arma::vec homotopySignature;
	
	bool interp_initiated;
	gsl_interp* workspace_x;
	gsl_interp* workspace_y;
	std::vector< double > getCoords(double tt);
	
	Traj_XYT(int nn);
	void _free(void);  // Needs to be called manually!!
};

//--------------------------------------------

static D3N2_MovingObstacleIn2D_Env* Active_Env_p;

class D3N2_MovingObstacleIn2D_Env
{
public:
	// It is important that we maintain only one instance of D3N2_MovingObstacleIn2D_Env at a time
	
	void parseEnvFile(const char* fName, bool vis=false);
	void parseObstacle(const YAML::Node& nObs);
	
	std::string problem_description;
	double tStart, tEnd;
	double loop_neginf_t, loop_posinf_t, conj_x, conj_y, conj_t;
	
	std::vector<GenericObstacle*> obstacles;
	std::vector<Traj_XYT> trajectories; // Outputed and/or provided in env file
	
	// Plotting / visualization
	DiscContTransform plotTransform;
	double plotTimestep;
	
	// Planning
	DiscContTransform planTransform;
	std::vector<double> startcoord, stopcoord;
	int exploreMaxCount;
	bool onlyJordan;
	std::vector<double> mins_planner, maxs_planner, cellsize_planner; // planner's dommain boundaries and cell-sizes
	
	D3N2_MovingObstacleIn2D_Env();
	~D3N2_MovingObstacleIn2D_Env();
};

// Functions for HomotopySearchProblem
bool isCoordFree(arma::vec coord);
arma::mat metricTensor_f(arma::vec coord);
double heuristics_f(arma::vec c1, arma::vec c2);

// =====================================================

#endif

