// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <fstream>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
#include <opencv/cv.h>
#include <opencv/cvaux.h>
#include <opencv/highgui.h> 

// Local includes
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "homotopy_NN_Astar.h"
#include "D3N2_MovingObstacleIn2D.h"

// ----------------------------------

void AddSingularityManifolds (D3N2_MovingObstacleIn2D_Env& myEnv, HomotopyProblem& myProb)
{
	std::vector< arma::vec > SingPts;
	double xs, ys;
	arma::vec thisCoord(2);
	for (int a=0; a<myEnv.obstacles.size(); a++)
	{
		if (myEnv.obstacles[a]->homotopySkeleton.size() < 2) // Obstacle to ignore for homotopy
			continue;
		
		SingPts.clear();
		double tt = myEnv.tStart;
		while (tt <= myEnv.tEnd)
		{
			thisCoord = myEnv.obstacles[a]->getSkeletonPos(tt);
			SingPts.push_back( _vArma(3, thisCoord[0], thisCoord[1], tt) );
			tt += myEnv.obstacles[a]->simplexSampleRate.eval(tt);
		}
		// point at positive infinity
		thisCoord = myEnv.obstacles[a]->getSkeletonPos(myEnv.loop_posinf_t);
		SingPts.push_back( _vArma(3, thisCoord[0], thisCoord[1], myEnv.loop_posinf_t) );
		// "conjugate" point
		SingPts.push_back( _vArma(3, myEnv.conj_x, myEnv.conj_y, myEnv.conj_t) );
		// point at negative infinity
		thisCoord = myEnv.obstacles[a]->getSkeletonPos(myEnv.loop_neginf_t);
		SingPts.push_back( _vArma(3, thisCoord[0], thisCoord[1], myEnv.loop_neginf_t) );
		
		myProb.addSingularityManifold( Create_1torus(SingPts) );
	}
}


arma::vec IntegrateOnTrajectory(Traj_XYT& traj, HomotopyProblem& myProb)
{
	std::vector< arma::vec > trajPts;
	for (int a=0; a<traj.n; a++)
		trajPts.push_back( _vArma(3, traj.x[a], traj.y[a], traj.t[a]) );
	SimplicesList theTrajManifold = Create_OpenCurve(trajPts);
	
	/* DEBUG 
	for (int a=0; a<myProb.singularityManifolds[0].s.size(); a++)
		myProb.singularityManifolds[0].s[a].print("\nSingularity simplex:");
	for (int a=0; a<theTrajManifold.s.size(); a++)
		theTrajManifold.s[a].print("\nIntegration simplex:"); */
	
	arma::vec intres = myProb.computeIntegrals(theTrajManifold);
	traj.homotopySignature = intres;
	return intres;
}

// -----------------------------------------

std::string CreateOutputFileWithTrajectories(const char* fName, 
								std::vector< std::vector< arma::vec > > trajs, std::vector< arma::vec > sigs)
{
	char outFname[1024], timestamp[256];
	strcpy(outFname, fName);
	time_t ltime = time(NULL);
	struct tm* Tm = localtime (&ltime);
	sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.out", 
							Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
	strcat(outFname, timestamp);
	
	std::string line;
	std::ifstream fenv(fName);
	std::ofstream fout(outFname);
	if (fout.is_open())
	{
		// Copy environment file to output file
		if (fenv.is_open())
		{
			while (fenv.good())
			{
				getline(fenv, line);
				fout << line << std::endl;
			}
			fenv.close();
		}
		// Print the trajectories
		fout << "\n\n# The following are added by the program\n" << std::endl;
		fout << "outputTrajectories:" << std::endl;
		for (int a=0; a<trajs.size(); a++)
		{
			fout << "  - homotopySignature: [";
			for (int b=0; b<sigs[a].n_elem; b++)
			{
				fout << sigs[a][b];
				if (b<sigs[a].n_elem-1)
					fout << ",";
			}
			fout << "]" << std::endl;
			fout << "    trajectory: [";
			for (int b=0; b<trajs[a].size(); b++)
			{
				fout << "[";
				for (int c=0; c<trajs[a][b].n_elem; c++)
				{
					fout << trajs[a][b][c];
					if (c<trajs[a][b].n_elem-1)
						fout << ",";
				}
				fout << "]";
				if (b<trajs[a].size()-1)
					fout << ",";
			}
			fout << "]" << std::endl;
		}
		fout.close();
	}
	
	std::string retFname = outFname;
	return retFname;
}


// ==========================================================

int main(int argc, char *argv[])
{
	printf("\n");
	D3N2_MovingObstacleIn2D_Env myEnv;
	char envFname[1024], movFname[1024], timestamp[1024];
	bool vis = false, savemov = false;
	
	if (argc < 2)
	{
		printf("Error: No argument provided.\n");
		exit(1);
	}
	else if ( !strcmp(argv[1], "-v") || !strcmp(argv[1], "-vo") ) // visualization & video output
	{
		vis = true;
		strcpy (envFname, argv[2]);
		if (!strcmp(argv[1], "-vo"))
		{
			savemov = true;
			strcpy(movFname, envFname);
			time_t ltime = time(NULL);
			struct tm* Tm = localtime (&ltime);
			sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", 
									Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
			strcat(movFname, timestamp);
		}
	}
	else
		strcpy (envFname, argv[1]);
	
	// Parse environment file or output file
	myEnv.parseEnvFile(envFname, vis);
	
	if (!vis)
	{
		std::cout << "Problem description: " << myEnv.problem_description << std::endl;
		
		// Initiate basic homotopy class problem
		int D=3, N=2;
		HomotopyProblem myProb(D, N);
		AddSingularityManifolds (myEnv, myProb);
		
		// -------------------------------------------------
		// Known trajectory - just evaluate the integral
		if (!myEnv.problem_description.compare("Check known trajectory") && myEnv.trajectories.size()==1)
		{
			arma::vec intres = IntegrateOnTrajectory(myEnv.trajectories[0], myProb);
			intres.print("The homotopy class signature for the given trajectory is");
		}
		// -------------------------------------------------
		// Explore homotopy classes
		if (!myEnv.problem_description.compare("Explore homotopy classes"))
		{
			HomotopySearchProblem searchProb;
			
			searchProb.dcTransform = myEnv.planTransform;
			searchProb.homotopyProb = &myProb;
			searchProb.start = _vArma( 3, myEnv.startcoord[0], myEnv.startcoord[1], myEnv.tStart);
			searchProb.stop = _vArma( 3, myEnv.stopcoord[0], myEnv.stopcoord[1], myEnv.tEnd);
			
			searchProb.isCoordFree = &isCoordFree;
			
			// List the Actions and their costs
			searchProb.forwardSteps.push_back(_vArma(3, myEnv.cellsize_planner[0], myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, -myEnv.cellsize_planner[0], myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, myEnv.cellsize_planner[0], -myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, -myEnv.cellsize_planner[0], -myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, myEnv.cellsize_planner[0], 0.0, myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, -myEnv.cellsize_planner[0], 0.0, myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, 0.0, myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, 0.0, -myEnv.cellsize_planner[1], myEnv.cellsize_planner[2]));
			searchProb.forwardSteps.push_back(_vArma(3, 0.0, 0.0, myEnv.cellsize_planner[2]));
			
			searchProb.metricTensor_f = &metricTensor_f;
			searchProb.heuristics_f = &heuristics_f;
			
			searchProb.MaxClassesToExplore = myEnv.exploreMaxCount;
			searchProb.onlyExploreJordan = myEnv.onlyJordan;
			
			printf("Performing search for %d Homotopy classes...\n", searchProb.MaxClassesToExplore);
			searchProb.init_and_plan();
			printf("Found %d trajectories.\n", searchProb.plannedTrajs.size());
			
			std::string outfname = CreateOutputFileWithTrajectories(envFname, searchProb.plannedTrajs, searchProb.plannedTrajSignatures);
			printf("Output written in: %s\n\n", outfname.c_str());
		}
		
	}
	else
	{
		IplImage* ipl_image_p;
		IplImage* ipl_image_p_processed;
		CvSize map_size;
		int pressedkey;
		double textAng;
		CvFont textFont;
		cvInitFont( &textFont, CV_FONT_HERSHEY_PLAIN, 1.0f, 1.0f, 0.0f, 2 );
		
		cvNamedWindow("D3N2_MovingObstacleIn2D", CV_WINDOW_AUTOSIZE);
		map_size.width = myEnv.plotTransform.disc_size[0];
		map_size.height = myEnv.plotTransform.disc_size[1];
		// Original image
		ipl_image_p = cvCreateImage(map_size, IPL_DEPTH_8U, 3);
		ipl_image_p->origin = 1;
		ipl_image_p->widthStep = ipl_image_p->width * 3;
		// Processed/scaled image
		ipl_image_p_processed = cvCreateImage(map_size, IPL_DEPTH_8U, 3);
		ipl_image_p_processed->origin = 1;
		ipl_image_p_processed->widthStep = ipl_image_p_processed->width * 3;
		//printf("Image: %d, %d : %d, %d, %d, %d\n", map_size.width, map_size.height, ipl_image_p->width, ipl_image_p->widthStep, ipl_image_p->height, ipl_image_p->align );
		
		CvVideoWriter* writer;
		if (savemov)
			writer = cvCreateVideoWriter ( movFname, CV_FOURCC('D','I','V','X'), 1.0/myEnv.plotTimestep, map_size);
		
		std::vector<int> APoint;
		char robNo[3];
		std::vector<double> APoint_d;
		std::vector< std::vector<CvPoint> > trajHist( myEnv.trajectories.size() );
		for (double tt = myEnv.tStart; tt <= myEnv.tEnd; tt += myEnv.plotTimestep)
		{
			printf("Drawing frame: t = %f\n", tt);
			// Clear image
			cvSet(ipl_image_p, CV_RGB(255.0,255.0,255.0));
			
			// Draw trajectory history
			for (int a=0; a<myEnv.trajectories.size(); a++)
			{
				CvScalar TRAJ_COLOR_CV = CV_RGB ( 	150.0+100.0*((double)((a*7)%17))/16.0, 
													150.0+100.0*((double)((a*11)%19))/18.0, 
													150.0+100.0*((double)((a*13)%23))/22.0 		);
				
				for (int b=0; b<((int)trajHist[a].size())-1; b++)
					cvLine(ipl_image_p, trajHist[a][b], trajHist[a][b+1], TRAJ_COLOR_CV, 2);
			}
			
			// Draw start and goal pos
			APoint = myEnv.plotTransform.cont2disc( myEnv.startcoord );
			cvRectangle(ipl_image_p, cvPoint(APoint[0]-1, APoint[1]-1), cvPoint(APoint[0]+1, APoint[1]+1), CV_RGB(0.0,0.0,255.0));
			APoint = myEnv.plotTransform.cont2disc( myEnv.stopcoord );
			cvRectangle(ipl_image_p, cvPoint(APoint[0]-1, APoint[1]-1), cvPoint(APoint[0]+1, APoint[1]+1), CV_RGB(0.0,255.0,0.0));
			
			// Draw obstacles
			for (int a=0; a<myEnv.obstacles.size(); a++)
				myEnv.obstacles[a]->drawInImage_cv(ipl_image_p, tt);
			
			// Draw current pos
			for (int a=0; a<myEnv.trajectories.size(); a++)
			{
				APoint_d = myEnv.trajectories[a].getCoords(tt);
				APoint = myEnv.plotTransform.cont2disc( APoint_d );
				cvRectangle(ipl_image_p, cvPoint(APoint[0]-1, APoint[1]-1), cvPoint(APoint[0]+1, APoint[1]+1), ROBOT_COLOR_CV);
				sprintf(robNo, "%d", a);
				textAng = 2.0*PI_CONST*a/myEnv.trajectories.size();
				cvPutText(ipl_image_p, robNo, cvPoint(APoint[0]+8*cos(textAng)-4, APoint[1]+8*sin(textAng)-5), &textFont, CV_RGB(200,50,80));
				trajHist[a].push_back(cvPoint(APoint[0], APoint[1]));
			}
			
			cvShowImage("D3N2_MovingObstacleIn2D", ipl_image_p);
			cvConvertImage(ipl_image_p, ipl_image_p_processed, CV_CVTIMG_FLIP);
			if (savemov)
				cvWriteFrame(writer, ipl_image_p_processed);
			pressedkey = cvWaitKey((int)(myEnv.plotTimestep*1000));
			
			// If 'c' was pressed, capture image
			if ( pressedkey=='c' || pressedkey=='C' )
			{
				char imgFname[1024], timestamp[1024];
				strcpy(imgFname, envFname);
				time_t ltime = time(NULL);
				struct tm* Tm = localtime (&ltime);
				sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_%f.png", 
										Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec, tt);
				strcat(imgFname, timestamp);
				// -------------------------
	
				cvSaveImage(imgFname, ipl_image_p);
				printf("Frame saved in: %s\n", imgFname);
			}
		}
		if (savemov)
			cvReleaseVideoWriter(&writer);
		cvWaitKey();
		cvReleaseImage(&ipl_image_p);
	}
	
	printf("\n");
}


