// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <fstream>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <GL/glut.h>

// Local includes
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "homotopy_NN_Astar.h"
#include "D3N2_StaticObstacleIn3D.h"


// LOOP

Obstacle_loop::Obstacle_loop(const YAML::Node& ObsNode)
{
	shape = "loop";
	
	if (ObsNode["homotopy"]) // can set this to "none" for not considering in homotopy class
		ObsNode["homotopy"] >> homotopy;
	else
		homotopy = "auto";
	
	muCoordinate thisCoord;
	for (int a=0; a<ObsNode["v"].size(); a++)
	{
		ObsNode["v"][a] >> thisCoord;
		skeleton.push_back( thisCoord );
	}
	ObsNode["radius"] >> rad;
}

bool Obstacle_loop::isPointInside(arma::vec coord, double tt)
{
	int b;
	arma::vec v1, v2, v2normalized;
	double v2mag, d, l, thisTimeRad = rad.eval(tt) + padding;
	std::vector< arma::vec > thisTimeVertices;
	
	for (int a=0; a<skeleton.size(); a++)
		thisTimeVertices.push_back( skeleton[a].eval(tt) );
	
	for (int a=0; a<skeleton.size(); a++)
	{
		v1 = coord - thisTimeVertices[a];
		// Check if coord is inside the sphere at vertex [a]
		if ( arma::norm(v1, 2) <= thisTimeRad )
			return true;
		// Check if coord is inside the cylinder made by vertices [a] and [b]
		b = (a < skeleton.size()-1) ? a+1 : 0;
		v2 = thisTimeVertices[b] - thisTimeVertices[a];
		v2mag = arma::norm(v2, 2);
		v2normalized = v2 / v2mag;
		d = arma::norm( armaCross(v1, v2normalized), 2 );
		if (d <= thisTimeRad)
		{
			l = arma::dot(v1, v2normalized);
			if ( l>=0 && l<=v2mag )
				return true;
		}
	}
	
	return false;
}

void Obstacle_loop::draw_gl(double tt)
{
	double thisTimeRad = rad.eval(tt);
	int b;
	std::vector< arma::vec > thisTimeVertices;
	
	for (int a=0; a<skeleton.size(); a++)
	{
		thisTimeVertices.push_back( skeleton[a].eval(tt) );
		// Drawing the spheres at the vertices
		GLUquadric* sphere_quad = gluNewQuadric();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
			glTranslated( thisTimeVertices[a][0], thisTimeVertices[a][1], thisTimeVertices[a][2] );
		SET_OBSTACLE_COLOR_GL;
			gluSphere(sphere_quad, thisTimeRad, 25, 25);
		glPopMatrix();
		gluDeleteQuadric(sphere_quad);
	}
	
	// Draw the cylinders
	for (int a=0; a<skeleton.size(); a++)
	{
		b = (a < skeleton.size()-1) ? a+1 : 0;
		SET_OBSTACLE_COLOR_GL;
		glDrawCylinderBetweenPoints ( thisTimeVertices[a], thisTimeVertices[b], thisTimeRad );
	}
}

// ==================================================================
// BOX

Obstacle_box::Obstacle_box(const YAML::Node& ObsNode)
{
	shape = "box";
	
	if (ObsNode.FindValue("homotopy")) // can set this to "none" for not considering in homotopy class
	{
		ObsNode["homotopy"] >> homotopy;
	}
	else
		homotopy = "none";
	
	ObsNode["v1"] >> v1;
	ObsNode["v2"] >> v2;
}

bool Obstacle_box::isPointInside(arma::vec coord, double tt)
{
	arma::vec v1a, v2a;
	v1a = v1.eval(tt);
	v2a = v2.eval(tt);
	
	for (int a=0; a<coord.n_elem; a++)
		if ( coord[a] <= __min(v1a[a], v2a[a])-padding  ||  coord[a] >= __max(v1a[a], v2a[a])+padding )
			return false;
	
	return true;
}

void Obstacle_box::draw_gl(double tt)
{
	arma::vec v1a, v2a;
	v1a = v1.eval(tt);
	v2a = v2.eval(tt);
	
	SET_OBSTACLE_COLOR_GL;
	glDrawBox (v1a, v2a);
}

// ==================================================================
// ==================================================================

void D3N2_StaticObstacleIn3D_Env::parseObstacle(const YAML::Node& nObs)
{
	GenericObstacle* thisGenericObstacle = NULL;
	
	if (nObs.FindValue("loop"))
	{
		Obstacle_loop* thisObstacle = new Obstacle_loop(nObs["loop"]);
		thisGenericObstacle = (GenericObstacle*)thisObstacle;
	}
	else if (nObs.FindValue("box"))
	{
		Obstacle_box* thisObstacle = new Obstacle_box(nObs["box"]);
		thisGenericObstacle = (GenericObstacle*)thisObstacle;
	} // else if
	
	// -------------------------------------------
	if (thisGenericObstacle)
	{
		thisGenericObstacle->parentEnv = this;
		
		if (nObs[thisGenericObstacle->shape].FindValue("padding"))
			nObs[thisGenericObstacle->shape]["padding"] >> thisGenericObstacle->padding;
		else
			thisGenericObstacle->padding = 0.0;
		
		obstacles.push_back( thisGenericObstacle );
	}
}

// ******************************************************************
// ==================================================================

muExpression::muExpression(std::string expr)
{
	p = new mu::Parser;
	t = new double;
	p->DefineVar("t", t);
	p->SetExpr(expr);
	s = expr;
}

void muExpression::init(std::string expr)
{
	p = new mu::Parser;
	t = new double;
	p->DefineVar("t", t);
	p->SetExpr(expr);
	s = expr;
}

double muExpression::eval(double tt)
{
	*t = tt;
	return ( p->Eval() );
}

void operator >> (const YAML::Node& node, muExpression& ex)
{
	std::string expr;
	node >> expr;
	ex.init(expr);
}

//-------------------------------------------

arma::vec muCoordinate::eval(double tt)
{
	arma::vec ret(coords.size());
	for (int a=0; a<coords.size(); a++)
		ret[a] = coords[a].eval(tt);
	return ret;
}

void operator >> (const YAML::Node& node, muCoordinate& co)
{
	// node must be a list.
	muExpression tmpexpr;
	co.coords.clear();
	for (int a=0; a<node.size(); a++)
	{
		node[a] >> tmpexpr;
		co.coords.push_back(tmpexpr);
	}
}

void operator >> (const YAML::Node& node, arma::vec& v)
{
	// node must be a list.
	arma::vec ret (node.size());
	for (int a=0; a<node.size(); a++)
		node[a] >> ret[a];
	v = ret;
}

// -------------------------------------------

void Traj_T::_alloc(int nn, int dd)
{
	n = nn;
	d = dd;
	double* arr;
	for (int a=0; a<d; a++)
	{
		arr = new double[n];
		coords.push_back( arr );
	}
	allocated = true;
}

void Traj_T::_free(void)
{
	for (int a=0; a<d; a++)
		delete[] coords[a];
}

void operator >> (const YAML::Node& node, Traj_T& traj)
{
	if (!traj.allocated)
		traj._alloc(node["trajectory"].size(), 3);
	
	if (node.FindValue("homotopySignature"))
		node["homotopySignature"] >> traj.homotopySignature;
	
	for (int a=0; a<node["trajectory"].size(); a++)
		for (int b=0; b<3; b++)
			node["trajectory"][a][b] >> traj.coords[b][a];
}

// ==================================================================

void D3N2_StaticObstacleIn3D_Env::parseEnvFile(const char* fName, bool vis)
{
	/* std::ifstream fin(fName);
	YAML::Parser parser(fin);
	YAML::Node doc;
	parser.GetNextDocument(doc); */
	YAML::Node doc = YAML::LoadFile(fName);
	
	// ----------------------------------------
	doc["problem"]["description"] >> problem_description;
	
	if (!problem_description.compare("Explore homotopy classes") || !problem_description.compare("Plan with homotopy constraint"))
	{
		doc["problem"]["startcoord"] >> startcoord;
		doc["problem"]["stopcoord"] >> stopcoord;
		
		if (doc["problem"].FindValue("maxcount"))
			doc["problem"]["maxcount"] >> exploreMaxCount;
		else
			exploreMaxCount = 0;
		
		if (doc["problem"].FindValue("onlyjordan"))
			doc["problem"]["onlyjordan"] >> onlyJordan;
		else
			onlyJordan = false;
			
		if (doc["problem"].FindValue("dontexplorenonjordan"))
			doc["problem"]["dontexplorenonjordan"] >> dontExploreNonjordanStates;
		else
			dontExploreNonjordanStates = false;
			
		if (doc["problem"].FindValue("allowedsignatures"))
			for (int a=0; a<doc["problem"]["allowedsignatures"].size(); a++)
			{
				arma::vec thisSignature;
				doc["problem"]["allowedsignatures"][a] >> thisSignature;
				allowed_signatures.push_back ( thisSignature );
			}
		
		if (doc["problem"].FindValue("blockedsignatures"))
			for (int a=0; a<doc["problem"]["blockedsignatures"].size(); a++)
			{
				arma::vec thisSignature;
				doc["problem"]["blockedsignatures"][a] >> thisSignature;
				blocked_signatures.push_back ( thisSignature );
			}
	}
	else if (!problem_description.compare("Check known trajectories"))
	{
		if (doc.FindValue("testTrajectories"))
			for (int a=0; a<doc["testTrajectories"].size(); a++)
			{
				Traj_T temptraj;
				doc["testTrajectories"][a] >> temptraj;
				trajectories.push_back(temptraj);
			}
	}
	
	// ----------------------------------------
	if (doc["planoptions"])
	{
		mins_planner.resize(3); maxs_planner.resize(3); cellsize_planner.resize(3); // XYZ
		for (int a=0; a<3; a++)
		{
			doc["planoptions"]["domain"][a][0] >> mins_planner[a];
			doc["planoptions"]["domain"][a][1] >> maxs_planner[a];
			doc["planoptions"]["cellsize"][a] >> cellsize_planner[a];
		}
		planTransform.init(mins_planner, maxs_planner, cellsize_planner);
		
		if (doc["planoptions"]["singularitymanifoldjiggle"])
			doc["planoptions"]["singularitymanifoldjiggle"] >> SingularityManifoldJiggle;
		else
			SingularityManifoldJiggle = 0.001;
			
		if (doc["planoptions"].FindValue("precomputeandreplan"))
			doc["planoptions"]["precomputeandreplan"] >> precomputeAndReplan;
		else
			precomputeAndReplan = false;
	}
	
	// Drawing options
	if (doc["drawoptions"])
	{
		doc["drawoptions"]["windowsize"][0] >> winsizeX;
		doc["drawoptions"]["windowsize"][1] >> winsizeY;
		doc["drawoptions"]["camerapos"] >> camerapos;
		doc["drawoptions"]["lookat"] >> lookat;
		doc["drawoptions"]["fieldofviewY"] >> fovY;
	}
	
	// Parsing obstacles
	for (int a=0; a<doc["obstacles"].size(); a++)
		parseObstacle(doc["obstacles"][a]);
	
	// ----------------------------------------
	// Trajectories
	if (doc["outputTrajectories"])
		for (int a=0; a<doc["outputTrajectories"].size(); a++)
		{
			Traj_T temptraj;
			doc["outputTrajectories"][a] >> temptraj;
			trajectories.push_back(temptraj);
		}
	
}

// ---------------
// Default values

D3N2_StaticObstacleIn3D_Env::D3N2_StaticObstacleIn3D_Env()
{
	winsizeX = 800;
	winsizeY = 800;
	camerapos = _vArma(3, 0.0, 0.0, -10.0);
	lookat = _vArma(3, 0.0, 0.0, 0.0);
	fovY = 40;
}


// ==================================================================
// Utility functions

void glDrawCylinderBetweenPoints ( arma::vec p1, arma::vec p2, double radius )
{
	GLUquadric* cylinder_quad = gluNewQuadric();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
		arma::vec zunit = _vArma(3,0.0,0.0,1.0);
		arma::vec v2 = p2 - p1;
		double v2mag = arma::norm(v2, 2);
		arma::vec v2normalized = v2 / v2mag;
		arma::vec v3 = armaCross( zunit, v2normalized );
		double ang = atan2( arma::norm(v3, 2) , arma::dot(zunit, v2normalized) );
		glTranslated( p1[0], p1[1], p1[2] );
		if ( arma::norm(v3, 2) == 0.0 )
			v3 = _vArma(3,1.0,0.0,0.0);
		glRotated( ang*180.0/PI_CONST, v3[0], v3[1], v3[2] );
		gluCylinder(cylinder_quad, radius, radius, v2mag, 25, 2);
	glPopMatrix();
	gluDeleteQuadric(cylinder_quad);
}

void glDrawBox ( arma::vec p1, arma::vec p2 )
{
	arma::vec signs = p1 - p2;
	
	glBegin(GL_POLYGON);
		glNormal3d(signs[0], 0.0, 0.0);
		glVertex3d(p1[0],p1[1],p1[2]); glVertex3d(p1[0],p2[1],p1[2]); glVertex3d(p1[0],p2[1],p2[2]); glVertex3d(p1[0],p1[1],p2[2]);
	glEnd();
	glBegin(GL_POLYGON);
		glNormal3d(0.0, signs[1], 0.0);
		glVertex3d(p1[0],p1[1],p1[2]); glVertex3d(p2[0],p1[1],p1[2]); glVertex3d(p2[0],p1[1],p2[2]); glVertex3d(p1[0],p1[1],p2[2]);
	glEnd();
	glBegin(GL_POLYGON);
		glNormal3d(0.0, 0.0, signs[2]);
		glVertex3d(p1[0],p1[1],p1[2]); glVertex3d(p2[0],p1[1],p1[2]); glVertex3d(p2[0],p2[1],p1[2]); glVertex3d(p1[0],p2[1],p1[2]);
	glEnd();
	glBegin(GL_POLYGON);
		glNormal3d(-signs[0], 0.0, 0.0);
		glVertex3d(p2[0],p2[1],p2[2]); glVertex3d(p2[0],p1[1],p2[2]); glVertex3d(p2[0],p1[1],p1[2]); glVertex3d(p2[0],p2[1],p1[2]);
	glEnd();
	glBegin(GL_POLYGON);
		glNormal3d(0.0, -signs[1], 0.0);
		glVertex3d(p2[0],p2[1],p2[2]); glVertex3d(p1[0],p2[1],p2[2]); glVertex3d(p1[0],p2[1],p1[2]); glVertex3d(p2[0],p2[1],p1[2]);
	glEnd();
	glBegin(GL_POLYGON);
		glNormal3d(0.0, 0.0, -signs[2]);
		glVertex3d(p2[0],p2[1],p2[2]); glVertex3d(p1[0],p2[1],p2[2]); glVertex3d(p1[0],p1[1],p2[2]); glVertex3d(p2[0],p1[1],p2[2]);
	glEnd();
}

