// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <fstream>
#include "yaml-cpp/yaml.h"
#include "muParser.h"
#include <GL/glut.h>
#include <opencv/cv.h>
#include <opencv/cvaux.h>
#include <opencv/highgui.h> 

// Local includes
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "homotopy_NN_Astar.h"
#include "D3N2_StaticObstacleIn3D.h"
#include "zpr.h"

// ----------------------------------

D3N2_StaticObstacleIn3D_Env myEnv;

// =============================================
// ---------------------------------------------

bool isCoordFree(arma::vec coord)
{
	if ( coord[0]<myEnv.mins_planner[0] || coord[1]<myEnv.mins_planner[1] || coord[2]<myEnv.mins_planner[2] || 
			coord[0]>myEnv.maxs_planner[0] || coord[1]>myEnv.maxs_planner[1] || coord[2]>myEnv.maxs_planner[2]  )
		return false;

	for (int a=0; a<myEnv.obstacles.size(); a++)
		if ( myEnv.obstacles[a]->isPointInside(coord) )
			return false;
			
	return true;
}

// ---------------------------------------------
/*
arma::mat metricTensor_f(arma::vec coord)
{
	double dx = coord[0] - Active_Env_p->stopcoord[0];
	double dy = coord[1] - Active_Env_p->stopcoord[1];
	if ( dx*dx + dy*dy < Active_Env_p->cellsize_planner[0] * Active_Env_p->cellsize_planner[1] ) // close to goal
		return ( arma::diagmat( _vArma(3, 1.0, 1.0, 1e-8) ) );
	else
		return ( arma::diagmat( _vArma(3, 1.0, 1.0, 1e-4) ) );
} */

// ---------------------------------------------

double heuristics_f(arma::vec c1, arma::vec c2)
{
	// NN-connected heuristics
	double dx = fabs(c1[0] - c2[0]);
	double dy = fabs(c1[1] - c2[1]);
	double dz = fabs(c1[2] - c2[2]);
	double mind = __min( __min(dx,dy), dz);
	double ddx = fabs(dx-mind);
	double ddy = fabs(dy-mind);
	double ddz = fabs(dz-mind);
	double dd1 = __max(ddx, ddy);
	double dd2 = (ddz==0.0) ? __min(ddx, ddy) : ddz;
	return ( fabs(dd1-dd2) + SQRT2_CONST*(__min(dd1,dd2)) + SQRT3_CONST*mind );
}

// ==============================================================

void AddSingularityManifolds (HomotopyProblem& myProb)
{
	std::vector< arma::vec > SingPts;
	double xs, ys;
	arma::vec jiggle(3);
	for (int a=0; a<myEnv.obstacles.size(); a++)
	{
		if (!myEnv.obstacles[a]->homotopy.compare("none"))
			continue;
		SingPts.clear();
		for (int b=0; b<myEnv.obstacles[a]->skeleton.size(); b++)
		{
			jiggle = myEnv.SingularityManifoldJiggle * _vArma(3, _RAND_11, _RAND_11, _RAND_11);
			SingPts.push_back( myEnv.obstacles[a]->skeleton[b].eval() + jiggle );
		}
		myProb.addSingularityManifold( Create_1torus(SingPts) );
	}
}


// -----------------------------------------------------------------

std::string CreateOutputFileWithTrajectories(const char* fName, 
								std::vector< std::vector< arma::vec > > trajs, std::vector< arma::vec > sigs)
{
	char outFname[1024], timestamp[256];
	strcpy(outFname, fName);
	time_t ltime = time(NULL);
	struct tm* Tm = localtime (&ltime);
	sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.out", 
							Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
	strcat(outFname, timestamp);
	
	std::string line;
	std::ifstream fenv(fName);
	std::ofstream fout(outFname);
	if (fout.is_open())
	{
		// Copy environment file to output file
		if (fenv.is_open())
		{
			while (fenv.good())
			{
				getline(fenv, line);
				fout << line << std::endl;
			}
			fenv.close();
		}
		// Print the trajectories
		fout << "\n\n# The following are added by the program\n" << std::endl;
		fout << "outputTrajectories:" << std::endl;
		for (int a=0; a<trajs.size(); a++)
		{
			fout << "  - homotopySignature: [";
			for (int b=0; b<sigs[a].n_elem; b++)
			{
				fout << sigs[a][b];
				if (b<sigs[a].n_elem-1)
					fout << ",";
			}
			fout << "]" << std::endl;
			fout << "    trajectory: [";
			for (int b=0; b<trajs[a].size(); b++)
			{
				fout << "[";
				for (int c=0; c<trajs[a][b].n_elem; c++)
				{
					fout << trajs[a][b][c];
					if (c<trajs[a][b].n_elem-1)
						fout << ",";
				}
				fout << "]";
				if (b<trajs[a].size()-1)
					fout << ",";
			}
			fout << "]" << std::endl;
		}
		fout.close();
	}
	
	std::string retFname = outFname;
	return retFname;
}

// ==============================================================
// --------------------------------
// These are for OpenGL display
void MainOpenGLIteration(void);
void KeyDownHandlerOpenGL(unsigned char key, int x, int y);

static GLfloat light_ambient[]  = { 0.0, 0.0, 0.0, 1.0 };
static GLfloat light_diffuse[]  = { 0.5, 0.5, 0.5, 1.0 };
static GLfloat light_specular[] = { 0.2, 0.2, 0.2, 1.0 };
static GLfloat light_position[] = { 0.0, 0.0, -10.0, 0.0 };

static GLfloat mat_ambient[]    = { 0.7, 0.7, 0.7, 1.0 };
static GLfloat mat_diffuse[]    = { 0.8, 0.8, 0.8, 1.0 };
static GLfloat mat_specular[]   = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat high_shininess[] = { 100.0 };
// --------------------------------

char envFname[1024];

int main(int argc, char *argv[])
{
	printf("\n");
	char movFname[1024], timestamp[1024];
	bool vis = false, savemov = false;
	
	if (argc < 2)
	{
		printf("Error: No argument provided.\n");
		exit(1);
	}
	else if ( !strcmp(argv[1], "-v") || !strcmp(argv[1], "-vo") ) // visualization & video output
	{
		vis = true;
		strcpy (envFname, argv[2]);
		if (!strcmp(argv[1], "-vo"))
		{
			savemov = true;
			strcpy(movFname, envFname);
			time_t ltime = time(NULL);
			struct tm* Tm = localtime (&ltime);
			sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.avi", 
									Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
			strcat(movFname, timestamp);
		}
	}
	else
		strcpy (envFname, argv[1]);
		
	// Parse environment file or output file
	myEnv.parseEnvFile(envFname, vis);
	
	if (!vis)
	{
		std::cout << "Problem description: " << myEnv.problem_description << std::endl;
		
		// Initiate basic homotopy class problem
		int D=3, N=2;
		HomotopyProblem myProb(D, N);
		myProb.epsabs_gsl = 1e-4; // Integration net precision per edge.
		myProb.workspaceLimit_gsl = 100;
		
		AddSingularityManifolds (myProb);
		
		// -------------------------------------------------
		// Explore homotopy classes
		if (!myEnv.problem_description.compare("Explore homotopy classes") 
				|| !myEnv.problem_description.compare("Plan with homotopy constraint"))
		{
			HomotopySearchProblem searchProb;
			
			searchProb.dcTransform = myEnv.planTransform;
			searchProb.homotopyProb = &myProb;
			searchProb.start = myEnv.startcoord;
			searchProb.stop = myEnv.stopcoord;
			
			searchProb.isCoordFree = &isCoordFree;
			
			// List the Actions and their costs
			for (int p=-1; p<=1; p++)
				for (int q=-1; q<=1; q++)
					for (int r=-1; r<=1; r++)
					{
						if (p==0 && q==0 && r==0)
							continue;
						searchProb.forwardSteps.push_back(_vArma(3,	(double)p*myEnv.cellsize_planner[0], 
																	(double)q*myEnv.cellsize_planner[1], 
																	(double)r*myEnv.cellsize_planner[2]  ) );
						searchProb.stepCosts.push_back( sqrt((double)(p*p + q*q + r*r)) );
					}
			
			searchProb.heuristics_f = &heuristics_f;
			
			searchProb.MaxClassesToExplore = myEnv.exploreMaxCount;
			searchProb.allowed_signatures = myEnv.allowed_signatures;
			searchProb.blocked_signatures = myEnv.blocked_signatures;
			
			searchProb.onlyExploreJordan = myEnv.onlyJordan;
			searchProb.discardNonjordanStates = myEnv.dontExploreNonjordanStates;
			
			searchProb.GraphDiameter = _round( 2.0 * __max( __max( myEnv.planTransform.disc_size[0], myEnv.planTransform.disc_size[1] ), 
																	myEnv.planTransform.disc_size[2] ) * SQRT3_CONST );
			
			searchProb.doReplan = myEnv.precomputeAndReplan;
			printf("Performing search for %d Homotopy classes...\n", searchProb.MaxClassesToExplore);
			searchProb.init_and_plan();
			printf("Found %d trajectories.\n", searchProb.plannedTrajs.size());
			
			std::string outfname = CreateOutputFileWithTrajectories(envFname, searchProb.plannedTrajs, searchProb.plannedTrajSignatures);
			printf("Output written in: %s\n\n", outfname.c_str());
		}
		else if (!myEnv.problem_description.compare("Check known trajectories"))
		{
			for (int a=0; a<myEnv.trajectories.size(); a++)
			{
				std::vector< arma::vec > trajPts(myEnv.trajectories[a].n);
				for (int b=0; b<myEnv.trajectories[a].n; b++)
					trajPts[b] = _vArma(3, myEnv.trajectories[a].coords[0][b], 
											myEnv.trajectories[a].coords[1][b], 
											myEnv.trajectories[a].coords[2][b] );
				SimplicesList thisTestTraj = Create_OpenCurve(trajPts);
				arma::vec res2 = myProb.computeIntegrals(thisTestTraj);
				char hedr[1024];
				sprintf(hedr, "\nHomotopy signature of trajectory %d:", a);
				res2.print(hedr);
			}
		}
	}
	else
	{
		// OpenGL visualization
		
		glutInit(&argc, argv);
		glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE); //glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
		glutInitWindowSize(myEnv.winsizeX, myEnv.winsizeY);
		int win = glutCreateWindow("D3N2_StaticObstacleIn3D");

		glShadeModel(GL_SMOOTH);
		glClearColor(1.0f,1.0f,1.0f,0.0f);
		
		//glViewport(-2.0, -2.0, 4.0, 4.0);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(myEnv.fovY, ((double)myEnv.winsizeX)/((double)myEnv.winsizeY), 0.1, 100.0);
		//glScalef(1.0f, -1.0f, 1.0f);
		
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glScalef(-1.0f, 1.0f, 1.0f);
		gluLookAt (	myEnv.camerapos[0], myEnv.camerapos[1], myEnv.camerapos[2], // camera location
					myEnv.lookat[0], myEnv.lookat[1], myEnv.lookat[2], // look at point
					0.0, 1.0, 0.0 ); // camera "up" direction

		glutDisplayFunc(MainOpenGLIteration);
		glutIdleFunc(MainOpenGLIteration);
		glutKeyboardUpFunc(KeyDownHandlerOpenGL);
		
		zprReferencePoint[0] = myEnv.lookat[0]; zprReferencePoint[1] = myEnv.lookat[1]; zprReferencePoint[2] = myEnv.lookat[2]; 
		zprInit();
		//zprSelectionFunc(drawAxes);     /* Selection mode draw function */
		//zprPickFunc(pick);              /* Pick event client callback   */
		
		// Setting up lights

		light_position[0] = myEnv.camerapos[0]; light_position[1] = myEnv.camerapos[1]; light_position[2] = myEnv.camerapos[2]; 
		glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
		glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
		glLightfv(GL_LIGHT0, GL_POSITION, light_position);

		glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
		glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
		glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glDepthFunc(GL_LESS);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_NORMALIZE);
		glEnable(GL_COLOR_MATERIAL);


		// Main loop
		glutMainLoop();
	}
}

// ------------------------------------

void MainOpenGLIteration(void)
{
	double tt=0.0;
	glClearDepth(1.0f);
	glClearColor(1.0f,1.0f,1.0f,0.0f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	// Draw the obstacles
	for (int a=0; a<myEnv.obstacles.size(); a++)
		myEnv.obstacles[a]->draw_gl(tt);
	
	// Draw start and end of trajectories
	glPointSize (6.0);
	glBegin(GL_POINTS);
	glColor3d( 0.0, 200.0, 0.0 );
	if (myEnv.startcoord.n_elem > 0)
		glVertex3d ( myEnv.startcoord[0], myEnv.startcoord[1], myEnv.startcoord[2] );
	glColor3d( 200.0, 0.0, 0.0 );
	if (myEnv.stopcoord.n_elem > 0)
		glVertex3d ( myEnv.stopcoord[0], myEnv.stopcoord[1], myEnv.stopcoord[2] );
	glEnd();
	
	// Draw the trajectories
	arma::vec p1, p2;
	for (int a=0; a<myEnv.trajectories.size(); a++)
	{
		glColor3d(	(0.2+0.8*((double)(((a+3)*7)%17))/16.0), 
					(0.1+0.4*((double)((a*11)%19))/18.0), 
					(0.2+0.8*((double)(((a+7)*13)%5))/4.0)  );
		glLineWidth (4.0);
		
		glDisable(GL_LIGHTING);
		glBegin(GL_LINE_STRIP);
		for (int b=0; b<myEnv.trajectories[a].n; b++)
		{
			glNormal3d(0.0, 0.0, -1.0);
			glVertex3d ( myEnv.trajectories[a].coords[0][b], myEnv.trajectories[a].coords[1][b], myEnv.trajectories[a].coords[2][b] );
		}
		glEnd();
		glEnable(GL_LIGHTING);
		
		/* for (int b=0; b<(myEnv.trajectories[a].n-1); b++)
		{
			p1 = _vArma( 3, myEnv.trajectories[a].coords[0][b], myEnv.trajectories[a].coords[1][b], myEnv.trajectories[a].coords[2][b] );
			p2 = _vArma( 3, myEnv.trajectories[a].coords[0][b+1], 
								myEnv.trajectories[a].coords[1][b+1], myEnv.trajectories[a].coords[2][b+1] );
			glDrawCylinderBetweenPoints ( p1, p2, 0.02 );
		} */
	}
	
	glutSwapBuffers();
}

// ----------------------------------

void KeyDownHandlerOpenGL(unsigned char key, int x, int y)
{
	if ( key=='c' || key=='C' )
	{
		char imgFname[1024], timestamp[1024];
		strcpy(imgFname, envFname);
		time_t ltime = time(NULL);
		struct tm* Tm = localtime (&ltime);
		sprintf(timestamp, "_%d-%02d-%02d_%02d-%02d-%02d_.png", 
								Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec);
		strcat(imgFname, timestamp);
		// -------------------------
		
		int WIDTH, HEIGHT, BIT_R, BIT_G, BIT_B;
		GLvoid* imageData;
		CvSize size;
		IplImage* ipl_image_p;
	
		WIDTH = glutGet(GLUT_WINDOW_WIDTH);
		HEIGHT = glutGet(GLUT_WINDOW_HEIGHT);
		BIT_R = sizeof(unsigned char);
		BIT_G = sizeof(unsigned char);
		BIT_B = sizeof(unsigned char);
		imageData = malloc(WIDTH*HEIGHT*(BIT_R+BIT_G+BIT_B)+1000); //Allocate memory for storing the image
		glReadPixels(0, 0, WIDTH, HEIGHT, GL_BGR, GL_UNSIGNED_BYTE, imageData); //Copy the image to the array imageData

		size.width = WIDTH;
		size.height = HEIGHT;
		ipl_image_p = cvCreateImageHeader(size, IPL_DEPTH_8U, 3);
		ipl_image_p->imageData = (char*)imageData;
		ipl_image_p->imageDataOrigin = ipl_image_p->imageData;
		ipl_image_p->origin = 1;
		
		cvSaveImage(imgFname, ipl_image_p);
		free(imageData);
		printf("View (%dx%d) saved in: %s\n", WIDTH, HEIGHT, imgFname);
	}
}

