
#include <stdio.h>
#include <math.h>
#include <vector>
#include <armadillo>
#include <stdlib.h>
#include <stdarg.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>

#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"


HomotopyTerms::HomotopyTerms(int d, int n)
{
	D = d;
	N = n;
	A_dm1 = D * pow(PI_CONST, (double)D/2.0) / gsl_sf_gamma(1.0 + (double)D/2.0);
	
	// Compute allPartitionIndices: part^{D-N} ({0,1,2,...,D-2}) (i.e. partition on D-1 elements)
	unsigned long long int posIndex, testIndex, maxPosIndex; // This limits the value of D to 8*sizeof(unsigned long long int)
	maxPosIndex = ((unsigned long long int)pow((double)2, (double)(D-1))) - ((unsigned long long int)1); // D-1 counts of 1's
	std::vector<int> pl, pr;
	int sign;
	bool notValid;
	APartition thisPart;
	for (posIndex=0; posIndex<maxPosIndex; posIndex++)
	{
		pl.clear();
		pr.clear();
		sign = 1;
		testIndex = posIndex;
		for (int a=0; a<D-1; a++)
		{
			if ( testIndex & ((unsigned long long int)1) ) // 'a' goes to left bin
			{
				sign = sign * (int)pow(-1, a-pl.size());
				pl.push_back(a);
			}
			else // 'a' goes to right bin
			{
				pr.push_back(a);
			}
			
			if ( pl.size()>D-N || pr.size()>(N-1) )
				break;
			testIndex = testIndex >> 1;
		}
		
		if ( pl.size()!=D-N || pr.size()!=(N-1) ) // TODO: Can make posIndex to jump here
			continue;
		
		thisPart.set(pl, pr, sign);
		allPartitionIndices.push_back(thisPart);
		leftPartitions.push_back(pl);
		rightPartitions.push_back(pr);
	}
}


// ==========================================================

HomotopyIntegration_Full::HomotopyIntegration_Full(HomotopyTerms t, BasicSimplex s, BasicSimplex m, 
																	double epsrel, double epsabs, size_t workspaceLimit)
{
	// Check if t and s are compatible
	if ( s.dim_ambient != t.D || s.dim_simplex != t.D-t.N || m.dim_ambient != t.D || m.dim_simplex != t.N-1 )
		exit(1);
	
	D = t.D;
	N = t.N;
	IntegrationTerms = t;
	SingularitySimplex = s;
	ManifoldSimplex = m;
	epsrel_gsl = epsrel;
	epsabs_gsl = epsabs;
	workspaceLimit_gsl = workspaceLimit;
}

double HomotopyIntegration_Full::G_k(double k, arma::vec y_singularity, arma::vec y_manifold)
{
	arma::vec x_singularity = SingularitySimplex.NormalizationMatrix * y_singularity + SingularitySimplex.NormalizationVector;
	arma::vec x_manifold = ManifoldSimplex.NormalizationMatrix * y_manifold + ManifoldSimplex.NormalizationVector;
	
	double diff, den = 0.0;
	for (int l=0; l<D; l++)
	{
		diff = x_manifold[l] - x_singularity[l];
		den += diff*diff;
	}
	den = pow(den, ((double)D)/2.0); // TODO: Can use a combination of integer power and sqrt.
	
	return ( (x_manifold[k]-x_singularity[k]) / (den * IntegrationTerms.A_dm1) );
}


Level1IntegrationCoefs HomotopyIntegration_Full::computeCoefs_int_G_k(arma::vec y_singularity, arma::vec y_manifold)
{
	Level1IntegrationCoefs coef;
	coef.aa=0.0;
	coef.bb=0.0;
	coef.cc=0.0;
	coef.pp.resize(D); // PROF OPTIMIZATION
	coef.qq.resize(D); // PROF OPTIMIZATION
	
	double m, n;
	for (int l=0; l<D; l++)
	{
		if (y_singularity.n_elem < D-N)
			m = -SingularitySimplex.NormalizationMatrix(l, D-N-1);
		else
			m = ManifoldSimplex.NormalizationMatrix(l, N-2);
	
		n = 0.0;
		if (y_manifold.n_elem > 0)
			n += arma::dot(ManifoldSimplex.NormalizationMatrix.submat(l,0,l,y_manifold.n_elem-1), y_manifold);
		n += ManifoldSimplex.NormalizationVector(l);
		if (y_singularity.n_elem > 0)
			n -= arma::dot(SingularitySimplex.NormalizationMatrix.submat(l,0,l,y_singularity.n_elem-1), y_singularity);
		n -= SingularitySimplex.NormalizationVector(l);
	
		coef.aa += m*m;
		coef.bb += 2.0*m*n;
		coef.cc += n*n;
		//coef.pp.push_back(m);
		//coef.qq.push_back(n);
		coef.pp[l] = m; // PROF OPTIMIZATION
		coef.qq[l] = n; // PROF OPTIMIZATION
	}

	return coef;
}

// -------------------------------------------------

double integrand_gsl(double x, void* params)
{
	double uLim, result, error;
	IntegrationState state = *((IntegrationState*)params); // Local instance
	
	if (state.iLevel_manifold == state.hp->N)
		state.iLevel_manifold--;
	else if (state.iLevel_manifold > 0)
	{
		state.iLevel_manifold--;
		state.localVarVals_manifold.push_back(x);
	}
	else if (state.iLevel_manifold==0)
	{
		state.iLevel_singularity--;
		state.localVarVals_singularity.push_back(x);
	}
	
	if ( ( state.iLevel_manifold == state.hp->N - 1  &&  state.iLevel_singularity == state.hp->D - state.hp->N ) ||
			( state.iLevel_manifold == 0  &&  state.iLevel_singularity == state.hp->D - state.hp->N ) )
		uLim = 1.0;
	else
		uLim = x;

	arma::vec y_singularity, y_manifold;

	switch (state.iLevel_singularity+state.iLevel_manifold)
	{
		case 0:  // Function evaluation.
		{
			y_singularity = Std2Arma_vec(state.localVarVals_singularity);
			y_manifold = Std2Arma_vec(state.localVarVals_manifold);
			
			double thisResultPart;
			result = 0.0;
			for (int k=0; k<state.hp->D; k++)
			{
				thisResultPart = 0.0;
				for (int rho=0; rho<state.hp->IntegrationTerms.allPartitionIndices.size(); rho++)
					thisResultPart += state.hp->IntegrationTerms.allPartitionIndices[rho].sign * 
											state.hp->SingularitySimplex.volScales[k][rho] * state.hp->ManifoldSimplex.volScales[k][rho];
				result += pow(-1, k) * state.hp->G_k(k, y_singularity, y_manifold) * thisResultPart;
			}
			result *= ( pow(-1, state.hp->D - state.hp->N) * 
							state.hp->SingularitySimplex.orientation * state.hp->ManifoldSimplex.orientation );
			error = 0.0;
		}
		break;
		// -----------------------
		
		case 1:
		{
			y_singularity = Std2Arma_vec(state.localVarVals_singularity);
			y_manifold = Std2Arma_vec(state.localVarVals_manifold);
			Level1IntegrationCoefs coefs = state.hp->computeCoefs_int_G_k(y_singularity, y_manifold);
			
			double thisResultPart, GkIntegral;
			result = 0.0;
			for (int k=0; k<state.hp->D; k++)
			{
				thisResultPart = 0.0;
				for (int rho=0; rho<state.hp->IntegrationTerms.allPartitionIndices.size(); rho++)
					thisResultPart += state.hp->IntegrationTerms.allPartitionIndices[rho].sign * 
											state.hp->SingularitySimplex.volScales[k][rho] * state.hp->ManifoldSimplex.volScales[k][rho];
				GkIntegral = analytic_Level1_Integration(0.0, uLim, coefs.pp[k], coefs.qq[k], coefs.aa, coefs.bb, coefs.cc, state.hp->D);
										// TODO: Can avoid repeated computations inside the above function for same a,b,c
				result += pow(-1, k) * GkIntegral * thisResultPart;
			}
			result *= ( pow(-1, state.hp->D - state.hp->N) * 
							state.hp->SingularitySimplex.orientation * state.hp->ManifoldSimplex.orientation 
									/ state.hp->IntegrationTerms.A_dm1 );
			error = 0.0;
		}
		break;
		// -----------------------
		
		default: // Recursive call for one level deeper integration
		{
			gsl_integration_workspace* w = gsl_integration_workspace_alloc(state.hp->workspaceLimit_gsl);
			gsl_function F;
			F.function = (double (*)(double, void*))&integrand_gsl;
			F.params = (void*)&state;
			
			//gsl_integration_qag (&F, 0.0, uLim, 
			//						state.hp->epsabs_gsl, state.hp->epsrel_gsl, state.hp->workspaceLimit_gsl, GSL_INTEG_GAUSS31, w, 
			//							&result, &error);
			//printf("Debug: %e, %e\n", state.hp->epsrel_gsl, state.hp->epsabs_gsl);
			gsl_integration_qags (&F, 0.0, uLim, 
									state.hp->epsabs_gsl, state.hp->epsrel_gsl, state.hp->workspaceLimit_gsl, w, 
										&result, &error);
			gsl_integration_workspace_free(w);
		}
		break;
		// -----------------------
	}
	
	return result;
}


double  HomotopyIntegration_Full::integrate(void)
{
	// Integrate (...) dy'[D-N-1] dy'[D-N-2] ... dy'[1] dy'[0]  dy[N-2] dy[N-3] ... dy[1] dy[0] 
	//                                                    (i.e. integrating over a (D-N)x(N-1) dimensional simplex)
	double result, error;
	IntegrationState state;
	state.hp = this;
	state.iLevel_singularity = D-N;
	state.iLevel_manifold = N;
	
	result = integrand_gsl(1.0, (void*)&state);
	
	integrationResult = result;
	return result;
}

// =========================================================

HomotopyProblem::HomotopyProblem(int d, int n)
{
	D = d;
	N = n;
	problemTerms = *(new HomotopyTerms(D, N));
	epsrel_gsl = 0.0; //1e-6;
	epsabs_gsl = 1e-10;
	workspaceLimit_gsl = 1000;
}

void HomotopyProblem::addSingularityManifold(SimplicesList m)
{
	for (int a=0; a<m.s.size(); a++)
		if ( !(m.s[a].foundTransformation && m.s[a].foundVolumescales) )
			m.s[a].preprocess(true, &(problemTerms.leftPartitions));
	
	singularityManifolds.push_back(m);
}

void HomotopyProblem::setIntegrationManifold(SimplicesList im)
{
	for (int a=0; a<im.s.size(); a++)
		if ( !(im.s[a].foundTransformation && im.s[a].foundVolumescales) )
			im.s[a].preprocess(true, &(problemTerms.rightPartitions));
	
	integrationManifold = im;
}

// ------------------------------------------------

#if INTEGRATION_ERROR_HANDLE
	static gslErrorInfo LastGslErrInfo;
	void gsl_err_handler_homotopy_tools_cpp(const char* reason, const char* file, int line, int gsl_errno)
	{
		LastGslErrInfo.wasErrorOnLastIntegration = true;
		LastGslErrInfo.reason = reason;
		LastGslErrInfo.file = file;
		LastGslErrInfo.line = line;
		LastGslErrInfo.gsl_errno = gsl_errno;
	}
#endif

arma::vec HomotopyProblem::computeIntegrals(void)
{
	arma::vec ret(singularityManifolds.size());
	
	HomotopyIntegration_Full* theIntegration = new HomotopyIntegration_Full;
	double integrationVal, thisIntVal;
	
	for (int n=0; n<singularityManifolds.size(); n++)
	{
		integrationVal = 0.0;
		for (int a=0; a<singularityManifolds[n].s.size(); a++)
			for (int b=0; b<integrationManifold.s.size(); b++)
			{
				#if HOMOTOPY_VERBOSE>0
					printf("Integrating: Singularity Manifold %d/%d , Singularity Simplex %d/%d , Candidate Simplex %d/%d: ", 
							n+1, singularityManifolds.size(), a+1, singularityManifolds[n].s.size(), 
							b+1, integrationManifold.s.size() );
				#endif
				delete theIntegration;
				theIntegration = new HomotopyIntegration_Full(problemTerms, singularityManifolds[n].s[a], integrationManifold.s[b],
																epsrel_gsl, epsabs_gsl, workspaceLimit_gsl);
				#if INTEGRATION_ERROR_HANDLE
					LastGslErrInfo.wasErrorOnLastIntegration = false;
					gsl_error_handler_t* old_handler_fp = gsl_set_error_handler( &gsl_err_handler_homotopy_tools_cpp );
				#endif
				thisIntVal = theIntegration->integrate();
				#if INTEGRATION_ERROR_HANDLE
					if (LastGslErrInfo.wasErrorOnLastIntegration)
					{
						printf("WARNING: Error detected in integration in homotopy_tools.cpp ::\n");
						singularityManifolds[n].s[a].print("Singularity simplex:");
						integrationManifold.s[b].print("Integration simplex:");
						old_handler_fp( LastGslErrInfo.reason, LastGslErrInfo.file, LastGslErrInfo.line, LastGslErrInfo.gsl_errno );
					}
					gsl_set_error_handler( old_handler_fp );
				#endif
				#if HOMOTOPY_VERBOSE>0
				    printf("Value = %f\n", thisIntVal);
				#endif
				integrationVal += thisIntVal;
			}
		ret[n] = integrationVal;
	}
	
	return ret;
}

arma::vec HomotopyProblem::computeIntegrals(SimplicesList integration_manifold)
{
	setIntegrationManifold(integration_manifold);
	return computeIntegrals();
}

arma::vec HomotopyProblem::computeIntegrals(BasicSimplex integration_simplex)
{
	SimplicesList newSimplicesList;
	newSimplicesList.s.push_back(integration_simplex);
	setIntegrationManifold(newSimplicesList);
	return computeIntegrals();
}

// --------------------------------------

void HomotopyProblem::print(void)
{
    printf("\nProblem details: \n\tD = %d, N = %d\n\n", D, N);
    printf("\tNumber of singularity manifolds: %d\n", singularityManifolds.size());
    for (int a=0; a<singularityManifolds.size(); a++) {
        printf("\t\tSingularity manifold %d:", a);
        singularityManifolds[a].printSimplices_compact();
    }
    if ( integrationManifold.s.size()==0 )
         printf("\n\tIntegration manifold not set!\n");
    else {
        printf("\n\tIntegration manifold:");
        integrationManifold.printSimplices_compact();
    }
    printf("\n");
}

