#ifndef HOMOTOPY_TOOLS_H
#define HOMOTOPY_TOOLS_H

// Includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <armadillo>

#include "topology_tools.h"
#include "subh_utils.h"

#define INTEGRATION_ERROR_HANDLE 0
#define HOMOTOPY_VERBOSE 0

/* Class declarations */

class APartition
{
public:
	std::vector<int> pl;
	std::vector<int> pr;
	int sign;
	void set(std::vector<int> l, std::vector<int> r, int s) { pl=l; pr=r; sign=s; };
};

// ------------------------------

class HomotopyTerms
{
public:

	// Constants
	int D, N;
	double A_dm1;
	//std::vector<int> setD;
	std::vector<APartition> allPartitionIndices; // All partitions of 0,1,2,...,(D-2) [i.e. D-1 elements] into D-N and N-1 parts.
	std::vector< std::vector<int> > leftPartitions, rightPartitions; // Just a copy of above. Used for BasicSimplex::preprocess
	
	// Constructor
	HomotopyTerms() {};
	HomotopyTerms(int d, int n);

};

// ------------------------------

class Level1IntegrationCoefs
{
public:
	double aa, bb, cc;
	std::vector<double> pp, qq;
};

class HomotopyIntegration_Full
{
public:

	// Basic information
	HomotopyTerms IntegrationTerms;
	BasicSimplex SingularitySimplex; // D-N dimensional simplex embedded in D dimensional ambient space
	BasicSimplex ManifoldSimplex;    // N-1 dimensional simplex embedded in D dimensional ambient space
									 //   Both should have volScales already computed.
	
	int D, N;
	double epsrel_gsl, epsabs_gsl;
	size_t workspaceLimit_gsl;

	// Constructors and initiation
	HomotopyIntegration_Full() {};
	HomotopyIntegration_Full(HomotopyTerms t, BasicSimplex s, BasicSimplex m, 
								double epsrel=1e-6, double epsabs=1e-10, size_t workspaceLimit=1000);
	
	// Helper functions for performing integration
	double G_k(double k, arma::vec y_singularity, arma::vec y_manifold);
	Level1IntegrationCoefs computeCoefs_int_G_k(arma::vec y_singularity, arma::vec y_manifold);
	
	// Main function for integration
	double integrate(void);
	double integrationResult;
};


class IntegrationState // Integration progress tracking and leveling variables
{
public:
	HomotopyIntegration_Full* hp;

	int iLevel_singularity; // Starts as D-N, and is reduced by 1 at each integration level deep.
	std::vector<double> localVarVals_singularity; // Vector of length (D-N - iLevel_singularity).
	
	int iLevel_manifold; // Starts as N-1, and is reduced by 1 at each integration level deep.
	std::vector<double> localVarVals_manifold; // Vector of length (D-N - iLevel_manifold).
};


// The function used by GSL
double integrand_gsl(double x, void* params);


// ------------------------------

class HomotopyProblem
{
public:
	// Maintains a collection of singularity manifolds and provides functions for computing integration
	int D, N;
	HomotopyTerms problemTerms;
	std::vector< SimplicesList > singularityManifolds;
	SimplicesList integrationManifold;
	
	HomotopyProblem(int d, int n);
	void addSingularityManifold(SimplicesList m);
	void setIntegrationManifold(SimplicesList im);
	
	// Integration parameters
	double epsrel_gsl, epsabs_gsl; // Defaults to 0.0, 1e-10
	size_t workspaceLimit_gsl; // Defaults to 1000
	
	arma::vec computeIntegrals(void);
	arma::vec computeIntegrals(SimplicesList integration_manifold);
	arma::vec computeIntegrals(BasicSimplex integration_simplex);
	
	void print(void);
};

// ======================================================================

#endif

