
#include <stdio.h>
#include <math.h>
#include <vector>
#include <armadillo>
#include <stdlib.h>
#include <stdarg.h>
#include <opencv/cv.h>
#include <opencv/cvaux.h>
#include <csignal>

#include "subh_utils.h"


// ============================================================
// Helper function

arma::vec Std2Arma_vec(std::vector<double> vin)
{
	arma::vec vout(vin.size());
	for (int a=0; a<vin.size(); a++)
		vout[a] = vin[a];
	return vout;
	// return ( arma::conv_to< arma::colvec >::from(vin) ); // PROF OPTIMIZATION
}

std::vector<double> Arma2Std_vec(arma::vec vin)
{
	std::vector<double> vout(vin.n_elem);
	for (int a=0; a<vin.n_elem; a++)
		vout[a] = vin[a];
	return vout;
}

// ------------------------------------------------

std::vector<int> set_k(int D, int k)
{
	std::vector<int> ret;
	for (int a=0; a<D; a++)
		if (a==k)
			continue;
		else
			ret.push_back(a);
	return ret;
}

arma::vec _vArma(int count, ... )
{
	arma::vec ret(count);
	va_list listptr;
	va_start(listptr, count); 
	for(int i=0; i<count; ++i)
		ret[i] = va_arg(listptr, double);
	return ret;
}

std::vector<double> _vStd_d(int count, ... )
{
	std::vector<double> ret(count);
	va_list listptr;
	va_start(listptr, count); 
	for(int i=0; i<count; ++i)
		ret[i] = va_arg(listptr, double);
	return ret;
}

std::vector<int> _vStd_i(int count, ... )
{
	std::vector<int> ret(count);
	va_list listptr;
	va_start(listptr, count); 
	for(int i=0; i<count; ++i)
		ret[i] = va_arg(listptr, int);
	return ret;
}


// =================================================

DiscContTransform::DiscContTransform(std::vector<double> mins, std::vector<double> maxs, double cs)
{
	if ( mins.size() != maxs.size() )
		return;
	
	dim = mins.size();
	cont_min = mins;
	cont_max = maxs;
	
	for (int a=0; a<dim; a++)
	{
		cellsize.push_back(cs);
		disc_size.push_back( (int)ceil((cont_max[a]-cont_min[a]) / cs) + 1 );
	}
}

void DiscContTransform::init(std::vector<double> mins, std::vector<double> maxs, double cs)
{
	if ( mins.size() != maxs.size() )
		return;
	
	dim = mins.size();
	cont_min = mins;
	cont_max = maxs;
	
	for (int a=0; a<dim; a++)
	{
		cellsize.push_back(cs);
		disc_size.push_back( (int)ceil((cont_max[a]-cont_min[a]) / cs) + 1 );
	}
}

DiscContTransform::DiscContTransform(std::vector<double> mins, std::vector<double> maxs, std::vector<double> cs)
{
	if ( mins.size() != maxs.size() )
		return;
	
	dim = mins.size();
	cont_min = mins;
	cont_max = maxs;
	cellsize = cs;
	
	for (int a=0; a<dim; a++)
		disc_size.push_back( (int)ceil((cont_max[a]-cont_min[a]) / cellsize[a] ) + 1 );
}

void DiscContTransform::init(std::vector<double> mins, std::vector<double> maxs, std::vector<double> cs)
{
	if ( mins.size() != maxs.size() )
		return;
	
	dim = mins.size();
	cont_min = mins;
	cont_max = maxs;
	cellsize = cs;
	
	for (int a=0; a<dim; a++)
		disc_size.push_back( (int)ceil((cont_max[a]-cont_min[a]) / cellsize[a] ) + 1 );
}

std::vector<double> DiscContTransform::disc2cont(std::vector<int> disc)
{
	std::vector<double> cont;
	for (int a=0; a<dim; a++)
		cont.push_back( (float)disc[a]*cellsize[a] + cont_min[a] );
	return cont;
}

arma::vec DiscContTransform::disc2cont_arma(std::vector<int> disc)
{
	arma::vec cont(dim);
	for (int a=0; a<dim; a++)
		cont[a] = (float)disc[a]*cellsize[a] + cont_min[a];
	return cont;
}

std::vector<int> DiscContTransform::cont2disc(std::vector<double> cont)
{
	std::vector<int> disc;
	for (int a=0; a<dim; a++)
		disc.push_back( _round( (cont[a] - cont_min[a]) / cellsize[a] ) );
	return disc;
}

std::vector<int> DiscContTransform::cont2disc_arma(arma::vec cont)
{
	std::vector<int> disc;
	for (int a=0; a<dim; a++)
		disc.push_back( _round( (cont[a] - cont_min[a]) / cellsize[a] ) );
	return disc;
}

// =================================================

arma::vec armaCross(arma::vec v1, arma::vec v2)
{
	arma::vec ret(3);
	ret[0] = v1[1]*v2[2] - v1[2]*v2[1];
	ret[1] = v1[2]*v2[0] - v1[0]*v2[2];
	ret[2] = v1[0]*v2[1] - v1[1]*v2[0];
	return ret;
}

arma::mat armaVertCast(arma::mat top, arma::mat bottom)
{
	if (top.n_cols != top.n_cols)
	{
		printf("ERROR: armaVertCast: number of columns must be the same.\n");
		exit(1);
	}
	int nCols = top.n_cols;
	int nRows_t = top.n_rows, nRows_b = bottom.n_rows;
	
	arma::mat ret = arma::zeros<arma::mat>(nRows_t+nRows_b, nCols);
	for (int b=0; b<nCols; b++)
	{
		for (int a=0; a<nRows_t; a++)
			ret(a, b) = top(a, b);
		for (int a=0; a<nRows_b; a++)
			ret(nRows_t+a, b) = bottom(a, b);
	}
	
	return ret;
}

// =================================================

double computePermSign(std::vector<int> p)
{
	arma::mat permMat = arma::zeros<arma::mat>(p.size(), p.size());
	try
	{
		for (int a=0; a<p.size(); a++)
			permMat(a, p[a]) = 1;
	}
	catch (...)
	{
		printf("ERROR in computePermSign: p.size()=%d: ", p.size());
		for (int a=0; a<p.size(); a++)
			printf("%d,", p[a]);
		printf("\n");
		raise(SIGABRT);
		//exit(1);
	}
	return (arma::det(permMat));
}

// -------------------------

std::string CreateTimeStampedDerivedFilename(std::string originalFname, std::string extension)
{
	char outFname[1024];
	time_t ltime = time(NULL);
	struct tm* Tm = localtime (&ltime);
	sprintf(outFname, "%s_%d-%02d-%02d_%02d-%02d-%02d_.%s", originalFname.c_str(),
							Tm->tm_year+1900, Tm->tm_mon+1, Tm->tm_mday, Tm->tm_hour, Tm->tm_min, Tm->tm_sec,
								extension.c_str() );
	std::string outFnameString (outFname);
	return outFnameString;
}

// ===============================================================

double analytic_Level1_Integration(double L, double U, double p, double q, double a, double b, double c, int D)
{
	// Computes the following integration:
	//   \int_L^U \frac{ p y + q }{ (a y^2 + b y + c)^{D/2} } dy
	//      Assumptions: b^2 - 4ac < 0, and D is a non-negative integer
	// See "analytic_integral.nb" and http://en.wikipedia.org/wiki/List_of_integrals_of_rational_functions for details.
	
	switch (D)
	{
		case 0: // \int_L^U  (p y + q) dy
		{
			return ( p*(U*U-L*L)/2.0 + q*(U-L) );
		}
		
		case 1: // \int_L^U \frac{ p y + q }{ \sqrt(a y^2 + b y + c) } dy
		{
			double s1 = sqrt(c + L*(b + a*L));
			double s2 = sqrt(c + U*(b + a*U));
			double sa = sqrt(a);
			double twoa = 2.0*a;
			return ( ( 2.0*sa*p*(s2 - s1) + 
					  (b*p - twoa*q) * ( log( ((b + twoa*L)/sa + 2.0*s1) / ((b + twoa*U)/sa + 2.0*s2) ) ) ) 
																										/ (twoa*sa) );
		}
	
		case 2: // \int_L^U \frac{ p y + q }{ a y^2 + b y + c } dy
		{
			double twoa = 2.0*a;
			double sr = sqrt(-b*b + 4*a*c);
			double d1 = b*p - twoa*q;
			return ( ( 2.0*d1*( atan((b + twoa*L)/sr) - atan((b + twoa*U)/sr) ) + 
						sr*p*( log( (c + U*(b + a*U)) / (c + L*(b + a*L)) ) )  ) / (twoa*sr) );
		}
		
		default: // Use recursion
		{
			double n = (double)D / 2.0;
			double recRes = analytic_Level1_Integration(L, U, 0, 1.0, a, b, c, D-2);
			double two_cp = 2.0*c*p;
			double two_aq = 2.0*a*q;
			return ( ( pow(c + L*(b + a*L), 1-n)*(-two_cp + two_aq*L + b*(-(L*p) + q)) + 
						pow(c + U*(b + a*U), 1-n)*(two_cp - b*(q - p*U) - two_aq*U) + 
							(-3 + 2*n)*(b*p - two_aq)*recRes )  /  ((b*b - 4*a*c)*(n-1)) );
		}
	}
}

// ========================================================

// Geometry operations
std::vector<arma::vec> RotateAndTranslate( std::vector<arma::vec> inPoints, 
											arma::vec RotAxisBase, arma::vec RotAxisApex, double RotAngle, 
											arma::vec Translate )
{
	arma::vec vertexCoord;
	std::vector<arma::vec> outPoints;
	double cosRotAng = cos(RotAngle);
	double sinRotAng = sin(RotAngle);
	double oneMinusCosRotAng = 1.0 - cosRotAng;
	arma::vec RotAxisVec = RotAxisApex - RotAxisBase;
	double RotAxisMag = arma::norm(RotAxisVec, 2);
	arma::vec RotAxisUnit = RotAxisVec / RotAxisMag;
	for (int a=0; a<inPoints.size(); a++)
	{
		vertexCoord = inPoints[a] - RotAxisBase;
		// Rodrigues' rotation formula
		if ( RotAngle != 0.0  &&  RotAxisMag > 0.0 )
			vertexCoord = vertexCoord*cosRotAng + armaCross(RotAxisUnit, vertexCoord)*sinRotAng 
								+ arma::dot(RotAxisUnit, vertexCoord)*oneMinusCosRotAng*RotAxisUnit;
		vertexCoord = vertexCoord + RotAxisBase + Translate;
		outPoints.push_back( vertexCoord );
	}
	return outPoints;
}

// =========================================================

arma::vec FindLineCoeffs( std::vector< arma::vec > s )
{
    arma::vec ret(3);
    ret[0] = s[1][1] - s[0][1];
    ret[1] = - ( s[1][0] - s[0][0] );
    ret[2] = -ret[0]*s[0][0] - ret[1]*s[0][1];
    return(ret);
}

arma::vec FindIntersectionBetweenSegments( std::vector< arma::vec > s1, std::vector< arma::vec > s2 )
{
    arma::vec c1, c2, intpt;
    arma::mat coefs(2,2);
    
    c1 = FindLineCoeffs(s1);
    c2 = FindLineCoeffs(s2);
    
    coefs[0] = c1[0]; coefs[1] = c2[0]; coefs[2] = c1[1]; coefs[3] = c2[1];
    intpt = - arma::inv(coefs) * _vArma(2, c1[2], c2[2]);
    
    if (intpt[0]<__min(s1[0][0],s1[1][0]) || intpt[0]>__max(s1[0][0],s1[1][0]) || intpt[1]<__min(s1[0][1],s1[1][1]) || intpt[1]>__max(s1[0][1],s1[1][1]) || 
        intpt[0]<__min(s2[0][0],s2[1][0]) || intpt[0]>__max(s2[0][0],s2[1][0]) || intpt[1]<__min(s2[0][1],s2[1][1]) || intpt[1]>__max(s2[0][1],s2[1][1]) )
        return (_vArma(1, 0.0));
    else
        return (intpt);
}

