#include <stdio.h>
#include <math.h>
#include <vector>
#include <armadillo>
#include <stdlib.h>
#include <stdarg.h>

#include "topology_tools.h"
#include "topology_aux.h"
#include "subh_utils.h"


// ========================================================
// 2-SPHERE
//    theta: latitude: -\pi/2 \leq \theta \leq \pi/2
//    phi: longitude: 0 \leq \phi \leq 2\pi

arma::vec Normal2Sphere_Vertex(double theta, double phi)
{
	return _vArma( 3, 
			cos(theta)*cos(phi), 
			cos(theta)*sin(phi), 
			sin(theta) );
}

SimplicesList Create_2Sphere(arma::mat transMat, arma::vec transVec, double Dtheta, double Dphi)
{
	const double eps = 1e-10;
	int countPhi = _round(2.0*PI_CONST/Dphi); // number of longitudes
	int dim_amb = transMat.n_rows;
	int offset_size;
	
	BasicSimplex* thisSimplex;
	double theta, phi;
	SimplicesList ret;
	
	for (int h=1; h>=-1; h-=2)
	{
		offset_size = ret.s.size();
		
		// hemesphere main
		for (theta=0.0; h*theta < PI_CONST/2.0-(Dtheta+eps); theta+=h*Dtheta)
		{
			for (phi=0.0; phi < 2.0*PI_CONST-eps; phi+=Dphi)
			{
				// first simplex
				thisSimplex = new BasicSimplex(2, dim_amb);
				thisSimplex->setVertex(0, transMat*Normal2Sphere_Vertex(theta, phi) + transVec );
				thisSimplex->setVertex(1, transMat*Normal2Sphere_Vertex(theta+h*Dtheta, phi) + transVec );
				thisSimplex->setVertex(2, transMat*Normal2Sphere_Vertex(theta+h*Dtheta, phi+Dphi) + transVec );
				ret.s.push_back(*thisSimplex);
				delete thisSimplex;
			
				// second simplex
				thisSimplex = new BasicSimplex(2, dim_amb);
				thisSimplex->setVertex(0, transMat*Normal2Sphere_Vertex(theta, phi) + transVec );
				thisSimplex->setVertex(1, transMat*Normal2Sphere_Vertex(theta, phi+Dphi) + transVec );
				thisSimplex->setVertex(2, transMat*Normal2Sphere_Vertex(theta+h*Dtheta, phi+Dphi) + transVec );
				ret.s.push_back(*thisSimplex);
				delete thisSimplex;
			
				// Set orientation
				if (ret.s.size()==2+offset_size)
				{
					if (offset_size>0)
						ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[1]) );
					ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
				}
				else if (phi==0.0)
				{
					ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2*(countPhi+1)]) );
					ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
				}
				else
				{
					ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-3]) );
					ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
				}
			}
		}
		
		// hemesphere cap
		theta = h*(PI_CONST/2.0 - Dtheta);
		for (phi=0.0; phi < 2.0*PI_CONST-eps; phi+=Dphi)
		{
			thisSimplex = new BasicSimplex(2, dim_amb);
			thisSimplex->setVertex(0, transMat*Normal2Sphere_Vertex(theta, phi) + transVec );
			thisSimplex->setVertex(1, transMat*Normal2Sphere_Vertex(theta+h*Dtheta, phi) + transVec );
			thisSimplex->setVertex(2, transMat*Normal2Sphere_Vertex(theta, phi+Dphi) + transVec );
			ret.s.push_back(*thisSimplex);
			delete thisSimplex;
		
			if (ret.s.size()==1+offset_size)
			{
				if (offset_size>0)
					ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[0]) );
			}
			else if (phi==0.0)
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2*countPhi-1]) );
			else
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
		}
	}
	
	return ret;
}



// ========================================================
// 2-TORUS
//    theta: 0 \leq \theta \leq 2\pi
//    phi: 0 \leq \phi \leq 2\pi

arma::vec Normal2Torus_Vertex(double aspect, double theta, double phi)
{
	return _vArma( 3, 
			(aspect + cos(phi))*cos(theta), 
			(aspect + cos(phi))*sin(theta), 
			sin(phi) );
}

SimplicesList Create_2Torus(arma::mat transMat, arma::vec transVec, double aspect, double Dtheta, double Dphi)
{
	const double eps = 1e-10;
	int countPhi = _round(2.0*PI_CONST/Dphi); // number of longitudes
	int dim_amb = transMat.n_rows;
	
	BasicSimplex* thisSimplex;
	double theta, phi;
	SimplicesList ret;
	
	for (theta=0.0; theta < 2.0*PI_CONST-eps; theta+=Dtheta)
	{
		for (phi=0.0; phi < 2.0*PI_CONST-eps; phi+=Dphi)
		{
			// first simplex
			thisSimplex = new BasicSimplex(2, dim_amb);
			thisSimplex->setVertex(0, transMat*Normal2Torus_Vertex(aspect, theta, phi) + transVec );
			thisSimplex->setVertex(1, transMat*Normal2Torus_Vertex(aspect, theta+Dtheta, phi) + transVec );
			thisSimplex->setVertex(2, transMat*Normal2Torus_Vertex(aspect, theta+Dtheta, phi+Dphi) + transVec );
			ret.s.push_back(*thisSimplex);
			delete thisSimplex;
		
			// second simplex
			thisSimplex = new BasicSimplex(2, dim_amb);
			thisSimplex->setVertex(0, transMat*Normal2Torus_Vertex(aspect, theta, phi) + transVec );
			thisSimplex->setVertex(1, transMat*Normal2Torus_Vertex(aspect, theta, phi+Dphi) + transVec );
			thisSimplex->setVertex(2, transMat*Normal2Torus_Vertex(aspect, theta+Dtheta, phi+Dphi) + transVec );
			ret.s.push_back(*thisSimplex);
			delete thisSimplex;
		
			// Set orientation
			if (ret.s.size()==2)
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
			else if (phi==0.0)
			{
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2*(countPhi+1)]) );
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
			}
			else
			{
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-3]) );
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
			}
		}
	}
	
	return ret;
}


