
#include <stdio.h>
#include <math.h>
#include <vector>
#include <armadillo>
#include <stdlib.h>
#include <stdarg.h>

#include "topology_tools.h"
#include "subh_utils.h"


BasicSimplex::BasicSimplex(int D_simplex, int D_ambient)
{
	dim_ambient = D_ambient;
	dim_simplex = D_simplex;
	arma::vec thisVertex(dim_ambient);
	for (int a=0; a<=dim_simplex; a++) // dim_simplex+1 number of vertices
		v.push_back(thisVertex);
	orientation = 1;
	foundTransformation = false;
	foundVolumescales = false;
}

// ------------------------------

void BasicSimplex::preprocess(bool findVolumeScales, std::vector< std::vector<int> >* volumeScalePartitions)
{
	//   v[0] is mapped to [0,0,...,0,0]; v[1] is mapped to [1,0,...,0]; v[2] is mapped to [1,1,...,0]; etc.		
	
	NormalizationVector = v[0];
	NormalizationMatrix.set_size(dim_ambient, dim_simplex);
	
	std::vector<arma::vec> shifted_v;
	arma::vec thisShiftedV;
	for (int a=0; a<=dim_simplex; a++)
	{
		thisShiftedV = v[a] - NormalizationVector;
		shifted_v.push_back(thisShiftedV);
	}
	
	// Creating the matrix
	for (int a=0; a<dim_ambient; a++)
	{
		// Create a row
		double theSum = 0.0;
		for (int b=1; b<=dim_simplex; b++)
		{
			NormalizationMatrix(a, b-1) = shifted_v[b][a] - theSum;
			theSum += NormalizationMatrix(a, b-1);
		}
	}
	
	foundTransformation = true;
	
	// ----------------------------------------
	
	if (findVolumeScales)
	{
		// Finds T in:   dx(rho0) dx(rho1) ... dx(rho{dim_simplex-1})  =  T  dy[dim_simplex-1] dy[dim_simplex-2] ... dy[1] dy[0]
		if (volScales.size() > 0)
			volScales.clear();
	
		// for seclection of dim_simplex elements from {0, 1, 2, ..., dim_ambient-1}, we find the volume scale.
		if (!volumeScalePartitions)
		{
			// TODO: Compute volumeScalePartitions: volumeScalePartitions[a] is a std::vector<int> of size dim_simplex
		}
		
		arma::mat thisSubMat(dim_simplex, dim_simplex);
		std::vector<int> set_without_k;
		std::vector<double> thisKvolScales;
		double detval, reversalSign = (double)pow(-1, _round((dim_simplex-1)/2));
		for (int k=0; k<dim_ambient; k++)  // TODO: There are some repetations of computation below. Can avoid that.
		{
			thisKvolScales.clear();
			set_without_k = set_k(dim_ambient, k);
			for (int a=0; a<volumeScalePartitions->size(); a++)
			{
				for (int b=0; b<dim_simplex; b++)
					thisSubMat.row(b) = NormalizationMatrix.row( set_without_k[ volumeScalePartitions->at(a)[b] ] );
				
				if (thisSubMat.n_rows == 0)
					detval = 1.0;
				else
					detval = arma::det(thisSubMat);
				
				thisKvolScales.push_back( reversalSign * detval ); // for Homotopy use.
			}
			volScales.push_back(thisKvolScales);
		}
		
		foundVolumescales = true;
	}
}


int BasicSimplex::setSameOrientation(BasicSimplex* anotherSimplex, double VertexSepThresh)
{
	// Assumption: anotherSimplex and this has no overlap except that they share exactly one (dim_ambient-1) dimentional face.
	// Returns 0 upon failure. Otherwise returns the orientation.
	
	if ( dim_ambient!=anotherSimplex->dim_ambient || dim_simplex!=anotherSimplex->dim_simplex )
	{
		printf("WARNING: Error in setting orientation. Implex dimensions do not agree!\n");
		return 0;
	}
	
	// Find the vertex correspondences
	std::vector<int> correspondences(dim_simplex+1, -1), correspondences_inv(dim_simplex+1, -1);
	for (int a=0; a<=dim_simplex; a++)
		for (int b=0; b<=dim_simplex; b++)
			if (arma::norm(v[a] - anotherSimplex->v[b], 2) < VertexSepThresh)
			{
				correspondences[a] = b;
				correspondences_inv[b] = a;
				break;
			}
			
	int freeVert, freeVert_inv;
	for (int a=0; a<=dim_simplex; a++)
	{
		if (correspondences[a] == -1)
			freeVert = a;
		if (correspondences_inv[a] == -1)
			freeVert_inv = a;
	}
	correspondences[freeVert] = freeVert_inv;
	correspondences_inv[freeVert_inv] = freeVert;
	
	double pSign = computePermSign(correspondences);
	if (fabs(fabs(pSign)-1.0) > 1e-6)
	{
		printf("WARNING: Error in setting orientation. Correspondences could not be determined!\n");
		return 0;
	}
	else
		orientation = anotherSimplex->orientation * (- _round(pSign));
		
	return (orientation);
}


void BasicSimplex::print(std::string header)
{
	printf("%s\nVertices:\n", header.c_str());
	for (int a=0; a<dim_ambient; a++)
	{
		for (int b=0; b<=dim_simplex; b++)
			printf("\t%f", v[b][a]);
		printf("\n");
	}
	printf("Orientation: %d\n", orientation);
}
// -------------------------------------------------

void SimplicesList::printSimplices_compact(void)
{
	printf("\n");
	for (int a=0; a<s.size(); a++)
	{
		printf("s{%d}=[", a+1);
		for (int b=0; b<=s[a].dim_simplex; b++)
		{
			printf("[");
			for (int c=0; c<s[a].dim_ambient; c++)
			{
				printf("%f", s[a].v[b][c]);
				if (c<s[a].dim_ambient-1)
					printf(",");
			}
			printf("]");
			if (b<s[a].dim_simplex)
				printf(";");
		}
		printf("]; ");
	}
	printf("\n");
}

// ==========================================================
// ==========================================================

SimplicesList Create_1torus(std::vector< arma::vec > vertices)
{
	SimplicesList ret;
	int count = vertices.size();
	int dim_amb = vertices[0].n_elem;
	
	int nxt;
	BasicSimplex* thisSimplex;
	for (int v=0; v<count; v++)
	{
		nxt = (v < count-1) ? v+1 : 0;
		thisSimplex = new BasicSimplex(1, dim_amb);
		thisSimplex->setVertex(0, vertices[v]);
		thisSimplex->setVertex(1, vertices[nxt]);
		ret.s.push_back(*thisSimplex);
		delete thisSimplex;
		if (v > 0)
			ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
	}
	
	return ret;
}

// ---------------------

SimplicesList Create_OpenCurve(std::vector< arma::vec > vertices)
{
	SimplicesList ret;
	int count = vertices.size();
	int dim_amb = vertices[0].n_elem;
	
	int nxt;
	BasicSimplex* thisSimplex;
	for (int v=0; v<count-1; v++)
	{
		nxt = v+1;
		thisSimplex = new BasicSimplex(1, dim_amb);
		thisSimplex->setVertex(0, vertices[v]);
		thisSimplex->setVertex(1, vertices[nxt]);
		ret.s.push_back(*thisSimplex);
		delete thisSimplex;
		if (v > 0)
			ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
	}
	
	return ret;
}

// ---------------------

SimplicesList Create_2torus(std::vector< std::vector< arma::vec > > vertices)
{
	int count1 = vertices.size();
	int count2 = vertices[0].size();
	int count = count1 * count2;
	int dim_amb = vertices[0][0].n_elem;
	
	SimplicesList ret;
	
	int nxt1, nxt2;
	BasicSimplex* thisSimplex;
	for (int v1=0; v1<count1; v1++)
	{
		nxt1 = (v1 < count1-1) ? v1+1 : 0;
		for (int v2=0; v2<count2; v2++)
		{
			nxt2 = (v2 < count2-1) ? v2+1 : 0;
			
			// first simplex
			thisSimplex = new BasicSimplex(2, dim_amb);
			thisSimplex->setVertex(0, vertices[v1][v2]);
			thisSimplex->setVertex(1, vertices[nxt1][v2]);
			thisSimplex->setVertex(2, vertices[nxt1][nxt2]);
			ret.s.push_back(*thisSimplex);
			delete thisSimplex;
			
			// second simplex
			thisSimplex = new BasicSimplex(2, dim_amb);
			thisSimplex->setVertex(0, vertices[v1][v2]);
			thisSimplex->setVertex(1, vertices[v1][nxt2]);
			thisSimplex->setVertex(2, vertices[nxt1][nxt2]);
			ret.s.push_back(*thisSimplex);
			delete thisSimplex;
			
			// Set orientation
			if (v1==0 && v2==0)
			{
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
			}
			else if (v2==0)
			{
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2*(count2+1)]) );
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-1]) );
			}
			else
			{
				ret.s[ret.s.size()-2].setSameOrientation( &(ret.s[ret.s.size()-3]) );
				ret.s[ret.s.size()-1].setSameOrientation( &(ret.s[ret.s.size()-2]) );
			}
		}
	}
	
	return ret;
}

