// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "yagsbpl_base.h"
#include "planners/A_star.h"
#include "homotopy_NN_Astar.h"
#include <gsl/gsl_errno.h>

homotopyNode::homotopyNode() 
{
	coords = new int [Active_HomotopySearchProblem_p->D];
	homotopy_signature = new int [Active_HomotopySearchProblem_p->S];
}

homotopyNode& homotopyNode::operator=(const homotopyNode& other)
{
	// We do not do self-assignment check since we do not destroy own allocation
	
	int a;
	for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
		coords[a] = other.coords[a];
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		homotopy_signature[a] = other.homotopy_signature[a];

	return *this;
}

homotopyNode::homotopyNode(const homotopyNode& other)
{
	int a;
	
	coords = new int [Active_HomotopySearchProblem_p->D];
	homotopy_signature = new int [Active_HomotopySearchProblem_p->S];
	
	for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
		coords[a] = other.coords[a];
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		homotopy_signature[a] = other.homotopy_signature[a];
}

// ----------------------------------------------------

bool homotopyNode::operator==(const homotopyNode& n)
{
	int a;
	for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
		if (coords[a] != n.coords[a])
			return false;
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		if ( abs(homotopy_signature[a] - n.homotopy_signature[a]) > Active_HomotopySearchProblem_p->HomotopySignatureCompareThresh )
			return false;
	return true;
}

bool homotopyNode::isBaseCoordsSame(homotopyNode& n)
{
	for (int a=0; a<Active_HomotopySearchProblem_p->D; a++)
		if (coords[a] != n.coords[a])
			return false;
	return true;
}

// ----------------------------------------------------

homotopyNode operator+(const homotopyNode& lhs, const homotopyNode& rhs)
{
	int a;
	homotopyNode ret = lhs;
	for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
		ret.coords[a] += rhs.coords[a];
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		ret.homotopy_signature[a] += rhs.homotopy_signature[a];
	return ret;
}

homotopyNode operator-(const homotopyNode& lhs, const homotopyNode& rhs)
{
	int a;
	homotopyNode ret = lhs;
	for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
		ret.coords[a] -= rhs.coords[a];
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		ret.homotopy_signature[a] -= rhs.homotopy_signature[a];
	return ret;
}

// -------------------

void homotopyNode::print(void)
{
	printf("(");
	std::vector<double> actualcoords = Active_HomotopySearchProblem_p->dcTransform.disc2cont 
											( Array2StdVec<int>(coords, Active_HomotopySearchProblem_p->D) );
	for (int a=0; a<Active_HomotopySearchProblem_p->D; a++)
		printf("%0.3f ", actualcoords[a]);
	printf(": ");
	for (int a=0; a<Active_HomotopySearchProblem_p->S; a++)
		printf("%f ", ((double)homotopy_signature[a] / (double)Active_HomotopySearchProblem_p->HomotopySignatureMultiplier) );
	printf(")");
}

// ===========================================================================================

unsigned int HomotopySearchProblem::shuffle(unsigned int key)
{
  key += (key << 12);
  key ^= (key >> 22);
  key += (key << 4);
  key ^= (key >> 9);
  key += ((!key) << 10);
  key ^= (key >> 2);
  key += (key << 7);
  key ^= (key >> 12);
  return key;
}

int HomotopySearchProblem::getHashBin(homotopyNode& n)
{
	unsigned int aggr=0;
	for (int a=0; a<Active_HomotopySearchProblem_p->D; a++)
		aggr += shuffle( (unsigned int)n.coords[a] ) << a;
	return ((int)(aggr & ((unsigned int)(hashTableSize-1)) ));
}

// ------------------------

bool HomotopySearchProblem::isAccessible(homotopyNode& n)
{
	// If Non-jordan state, and we need to discard it.
	if (discardNonjordanStates)
		for (int a=0; a<Active_HomotopySearchProblem_p->S; a++)
			if ( abs(n.homotopy_signature[a]) > HomotopySignatureMultiplier)
				return false;
	// First search for coordinate match in planner's hash table
	int hBin = getHashBin(n);
	for (int a=0; a<planner->GenericPlannerInstance.hash->HashTable[hBin].size(); a++)
	{
		if ( n.isBaseCoordsSame( planner->GenericPlannerInstance.hash->HashTable[hBin][a]->n ) && 
								planner->GenericPlannerInstance.hash->HashTable[hBin][a]->initiated ) // found match
			return ( planner->GenericPlannerInstance.hash->HashTable[hBin][a]->plannerVars.accessible );
	}
	// If not found, invoke higher-level function
	arma::vec contcoords = dcTransform.disc2cont_arma( Array2StdVec<int>(n.coords, Active_HomotopySearchProblem_p->D) );
	return (isCoordFree(contcoords));
}

// ------------------------

void HomotopySearchProblem::getSuccessors(homotopyNode& n, std::vector<homotopyNode>* s, std::vector<int>* costs)
{
	int a, b, c;
	homotopyNode thisNode;
	double homotopy_component = 0.0;
	arma::mat metricT;
	s->clear(); // Planner is supposed to clear these. Still, just in case!
	costs->clear();

	// Check if base node exists in hash
	int hBin = getHashBin(n);
	for (a=0; a<planner->GenericPlannerInstance.hash->HashTable[hBin].size(); a++)
	{
		if ( n.isBaseCoordsSame( planner->GenericPlannerInstance.hash->HashTable[hBin][a]->n ) && 
						planner->GenericPlannerInstance.hash->HashTable[hBin][a]->initiated && 
								planner->GenericPlannerInstance.hash->HashTable[hBin][a]->successors.size() > 0 )
		{
			// printf("getSuccessors: Found successors from hash table.\n");
			for (b=0; b<planner->GenericPlannerInstance.hash->HashTable[hBin][a]->successors.size(); b++)
			{
				thisNode = n + 
							( planner->GenericPlannerInstance.hash->HashTable[hBin][a]->successors.getLinkSearchGraphNode(b)->n 
												- planner->GenericPlannerInstance.hash->HashTable[hBin][a]->n );
				s->push_back( thisNode );
				costs->push_back( planner->GenericPlannerInstance.hash->HashTable[hBin][a]->successors.getLinkCost(b) );
			}
			return;
		}
	}

	// If base node does not exist in hash, compute successors
	arma::vec contcoords = dcTransform.disc2cont_arma( Array2StdVec<int>(n.coords, Active_HomotopySearchProblem_p->D) );
	std::vector<arma::vec> actions;
	
	if ( forwardSteps_f )
		actions = forwardSteps_f(contcoords);
	else
		actions = forwardSteps;
			
	// check if "n" is itself a successor of each of the successor nodes
	bool foundInverseMatch = false;
	for (a=0; a<actions.size(); a++)
	{
		// Set the base coordinates of thisNode equal to the base coordinates of successor 'a'
		StdVec2AllocatedArray<int>( dcTransform.cont2disc_arma( contcoords + actions[a] ), thisNode.coords ); 
		
		// THE FOLLOWING WAS CAUSING PROBLEMS.
		// We check the successors of thisNode in hash table to see if any is pointing back to "n"
		hBin = getHashBin(thisNode);
		foundInverseMatch = false;
		for (b=0; b<planner->GenericPlannerInstance.hash->HashTable[hBin].size(); b++)
		{
			if ( thisNode.isBaseCoordsSame( planner->GenericPlannerInstance.hash->HashTable[hBin][b]->n ) && 
							planner->GenericPlannerInstance.hash->HashTable[hBin][b]->initiated && 
									planner->GenericPlannerInstance.hash->HashTable[hBin][b]->successors.size() > 0 )
			{
				// Found successors of thisNode! Check if any of them has same base coord as "n"
				for (c=0; c<planner->GenericPlannerInstance.hash->HashTable[hBin][b]->successors.size(); c++)
				{
					if ( n.isBaseCoordsSame( planner->GenericPlannerInstance.hash->HashTable[hBin][b]->successors.getLinkSearchGraphNode(c)->n ) )
					{
						// Found match! We can now derive the homotopy signature of thisNode from the match.
						thisNode = n + 
										( planner->GenericPlannerInstance.hash->HashTable[hBin][b]->n - 
												planner->GenericPlannerInstance.hash->HashTable[hBin][b]->successors.getLinkSearchGraphNode(c)->n );
						foundInverseMatch = true;
						break;
					}
				}
			}
			if (foundInverseMatch)
				break;
		}
		
		if (foundInverseMatch)
		{
			// printf("getSuccessors: Found inverse successors from hash table.\n");
			s->push_back( thisNode );
			
			// Computing costs
			if (homotopy_weight_in_cost > 0)
				homotopy_component = HomotopyCostComponent(thisNode, n);
				
			if ( actionCost_f )
				costs->push_back( _round( CostMultiplier * (actionCost_f(contcoords, actions[a])+homotopy_component) ) );
			else if ( metricTensor_f )
			{
				metricT = metricTensor_f( contcoords + actions[a]/2.0 );
				costs->push_back( _round( CostMultiplier * (sqrt( as_scalar( arma::trans(actions[a])*metricT*actions[a] ) ) + homotopy_component) ) );
			}
			else
				costs->push_back( _round( CostMultiplier * (stepCosts[a]+homotopy_component) ) );
		}
		else
		{
			// compute the homotopy signature using integration
			SimplicesList segment;
			BasicSimplex segmentSimplex(1, Active_HomotopySearchProblem_p->D);
			segmentSimplex.setVertex(0, dcTransform.disc2cont_arma( Array2StdVec<int>(n.coords, Active_HomotopySearchProblem_p->D) ) );
			segmentSimplex.setVertex(1, 
							dcTransform.disc2cont_arma( Array2StdVec<int>(thisNode.coords, Active_HomotopySearchProblem_p->D) ) );
			segment.s.push_back( segmentSimplex );
			//segmentSimplex.print("Now integrating over: segmentSimplex = ");
			
			Active_HomotopySearchProblem_p->wasErrorOnLastIntegration = false;
			arma::vec homotopy_diff_multiplied = HomotopySignatureMultiplier * homotopyProb->computeIntegrals(segment);
			if (Active_HomotopySearchProblem_p->wasErrorOnLastIntegration)
			{
				printf("WARNING: Encountered problem in numerical integration for an edge. Skipping it! (%s)\n", 
								Active_HomotopySearchProblem_p->errorReasonOnLastIntegration);
			}
			else
			{
				// Computing successor
				for (b=0; b<Active_HomotopySearchProblem_p->S; b++)
					thisNode.homotopy_signature[b] = n.homotopy_signature[b] + _round( homotopy_diff_multiplied[b] );
				s->push_back( thisNode );
				
				// Computing costs
				if (homotopy_weight_in_cost > 0)
					homotopy_component = HomotopyCostComponent(thisNode, n);
					
				if ( actionCost_f )
					costs->push_back( _round( CostMultiplier * (actionCost_f(contcoords, actions[a])+homotopy_component) ) );
				else 
				if ( metricTensor_f )
				{
					metricT = metricTensor_f( contcoords + actions[a]/2.0 );
					costs->push_back( _round( CostMultiplier * (sqrt( as_scalar( arma::trans(actions[a])*metricT*actions[a] ) ) + homotopy_component) ) );
				}
				else
					costs->push_back( _round( CostMultiplier * (stepCosts[a]+homotopy_component) ) );
			}
			
		}
	}
	
}

// ------------------------

int HomotopySearchProblem::getHeuristics(homotopyNode& n1, homotopyNode& n2)
{
	if (heuristics_f)
	{
		arma::vec contcoords1 = dcTransform.disc2cont_arma( Array2StdVec<int>(n1.coords, Active_HomotopySearchProblem_p->D) );
		arma::vec contcoords2 = dcTransform.disc2cont_arma( Array2StdVec<int>(n2.coords, Active_HomotopySearchProblem_p->D) );
		double homotopy_component = 0.0;
		if (homotopy_weight_in_cost > 0)
			homotopy_component = homotopy_weight_in_cost * HomotopySearchProblem::HomotopyCostComponent(n1, n2);
		// n1.print(); n2.print(); printf(" -- %f \n", homotopy_component);
		return ( floor( CostMultiplier * ( heuristics_f( contcoords1, contcoords2 ) + homotopy_component ) ) );
	}
	else
		return 0;
}

double HomotopySearchProblem::HomotopyCostComponent(homotopyNode& n1, homotopyNode& n2)
{
	arma::vec htpy_diff(Active_HomotopySearchProblem_p->S);
	for (int a=0; a<Active_HomotopySearchProblem_p->S; a++)
		htpy_diff[a] = ((double)(n1.homotopy_signature[a]-n2.homotopy_signature[a])) / ((double)HomotopySignatureMultiplier);
	return (homotopy_weight_in_cost * arma::norm(htpy_diff, 2) );
}

// ------------------------

bool HomotopySearchProblem::storePath(homotopyNode& n)
{
	int a, b;
	if ( MaxClassesToExplore > 0  &&  storedPathsCount < MaxClassesToExplore )
	{
		if (stopNode->isBaseCoordsSame(n))
		{
			if (onlyExploreJordan)
				for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
					if ( abs(n.homotopy_signature[a]) > HomotopySignatureMultiplier)
						return false;
						
			if (allowed_homotopy_signatures.size() > 0)
			{
				bool thisClass, inAllowed = false;
				for (a=0; a<allowed_homotopy_signatures.size(); a++)
				{
					thisClass = true;
					for (b=0; b<Active_HomotopySearchProblem_p->S; b++)
					{
						thisClass = thisClass & ( abs(n.homotopy_signature[b]-allowed_homotopy_signatures[a][b])
																						<= HomotopySignatureCompareThresh );
						if (!thisClass) break;
					}
					if (thisClass)
					{
						inAllowed = true;
						break;
					}
				}
				if (!inAllowed)
					return false;
			}
			
			if (blocked_homotopy_signatures.size() > 0)
			{
				bool thisClass;
				for (a=0; a<blocked_homotopy_signatures.size(); a++)
				{
					thisClass = true;
					for (b=0; b<Active_HomotopySearchProblem_p->S; b++)
					{
						thisClass = thisClass & ( abs(n.homotopy_signature[b]-blocked_homotopy_signatures[a][b])
																						<= HomotopySignatureCompareThresh );
						if (!thisClass) break;
					}
					if (thisClass)
						return false;
				}
			}
			
			storedPathsCount++;
			/* printf("* Stored a path: %d in total. ", storedPathsCount);
			#if VIEW_PROGRESS_A_STAR
				float timediff = ((float)(clock() - planner->startclock)) / ((float)CLOCKS_PER_SEC);
				printf("States expanded: %d. Time elapsed: %f s. ", 
							planner->expandcount, ((timediff>=0.0) ? timediff : difftime(time(NULL), planner->startsecond)));
			#endif
			printf("\n"); 
			printf("** %d, %f \n", Active_HomotopySearchProblem_p->S, homotopy_weight_in_cost); */
			if (homotopy_weight_in_cost > 0.0)
				for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
					planner->GraphDescriptor->TargetNode.homotopy_signature[a] = 
									_round( allowed_signatures[storedPathsCount][a] * HomotopySignatureMultiplier );
			/*planner->GraphDescriptor->TargetNode.print(); 
			printf(" - %f, %f - ", allowed_signatures[0][0], allowed_signatures[1][0]);
			printf("\n"); */
			
			return true;
		}
	}
	
	return false;
}

bool HomotopySearchProblem::stopSearch(homotopyNode& n)
{
	int a, b;
	
	if ( ( (MaxClassesToExplore > 0 && storedPathsCount >= MaxClassesToExplore-1) || 
				( MaxClassesToExplore==0 && (allowed_homotopy_signatures.size() > 0 || blocked_homotopy_signatures.size() > 0 ) ) )
			 &&  stopNode->isBaseCoordsSame(n) )
	{
		if (onlyExploreJordan)
			for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
				if ( abs(n.homotopy_signature[a]) > HomotopySignatureMultiplier)
					return false;
					
		if (allowed_homotopy_signatures.size() > 0)
		{
			bool thisClass, inAllowed = false;
			for (a=0; a<allowed_homotopy_signatures.size(); a++)
			{
				thisClass = true;
				for (b=0; b<Active_HomotopySearchProblem_p->S; b++)
				{
					thisClass = thisClass & ( abs(n.homotopy_signature[b]-allowed_homotopy_signatures[a][b])
																					<= HomotopySignatureCompareThresh );
					if (!thisClass) break;
				}
				if (thisClass)
				{
					inAllowed = true;
					break;
				}
			}
			if (!inAllowed)
				return false;
		}
		
		if (blocked_homotopy_signatures.size() > 0)
		{
			bool thisClass;
			for (a=0; a<blocked_homotopy_signatures.size(); a++)
			{
				thisClass = true;
				for (b=0; b<Active_HomotopySearchProblem_p->S; b++)
				{
					thisClass = thisClass & ( abs(n.homotopy_signature[b]-blocked_homotopy_signatures[a][b])
																					<= HomotopySignatureCompareThresh );
					if (!thisClass) break;
				}
				if (thisClass)
					return false;
			}
		}
		
		storedPathsCount++;
		/* printf("* Stored a path: %d in total. ", storedPathsCount);
		#if VIEW_PROGRESS_A_STAR
			float timediff = ((float)(clock() - planner->startclock)) / ((float)CLOCKS_PER_SEC);
			printf("States expanded: %d. Time elapsed: %f s. ", 
					planner->expandcount, ((timediff>=0.0) ? timediff : difftime(time(NULL), planner->startsecond)));
		#endif
		printf("Stopping further search!\n"); */
		return true;
	}
	
	return false;
}

// =========================================

HomotopySearchProblem::HomotopySearchProblem()
{ 
	// homotopy problem options
	HomotopySignatureMultiplier = 1000000;   // Integration should be accurate at least upto
	HomotopySignatureCompareThresh = 10000;  //        HomotopySignatureCompareThresh / HomotopySignatureMultiplier
	GraphDiameter = -1;
	MaxClassesToExplore = 0;
	onlyExploreJordan = false;
	discardNonjordanStates = false;
	
	// Planner options
	CostMultiplier = 100000; 
	hashTableSize = 32*1024; 
	forwardSteps_f = NULL;
	metricTensor_f = NULL;
	actionCost_f = NULL;
	heuristics_f = NULL;
	homotopy_weight_in_cost = -1.0;
	doReplan = false;
	A_star_suboptimality_eps = 1.0;
	A_star_heap_key_count = 20;
	A_star_progress_display_interval = 1000;
	
	storedPathsCount = 0;
	Active_HomotopySearchProblem_p = this;
}

void HomotopySearchProblem::init_and_plan(void)
{
	Active_HomotopySearchProblem_p->D = homotopyProb->D;
	Active_HomotopySearchProblem_p->S = homotopyProb->singularityManifolds.size();
	gsl_set_error_handler( &gsl_err_handler );
	
	// Compute HomotopySignatureCompareThresh if GraphDiameter is set.
	if (GraphDiameter>0)
	{
		HomotopySignatureCompareThresh = _round( HomotopySignatureMultiplier * homotopyProb->epsabs_gsl * GraphDiameter );
		if (HomotopySignatureCompareThresh < 100)
			printf("WARNING: HomotopySignatureCompareThresh = %ld. The value may be too small.\n", HomotopySignatureCompareThresh);
		else if (HomotopySignatureCompareThresh > HomotopySignatureMultiplier/100)
			printf("WARNING: HomotopySignatureCompareThresh = %ld. The value may be too large. (HomotopySignatureMultiplier = %ld)\n",
									HomotopySignatureCompareThresh, HomotopySignatureMultiplier);
	}
	
	// Set variables - start and stop nodes in discretized coordinates
	int a;
	
	startNode = new homotopyNode;
	//for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
	StdVec2AllocatedArray<int>( dcTransform.cont2disc_arma( start ), startNode->coords );
	for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
		startNode->homotopy_signature[a] = 0;
	
	stopNode = new homotopyNode;
	//for (a=0; a<Active_HomotopySearchProblem_p->D; a++)
	StdVec2AllocatedArray<int>( dcTransform.cont2disc_arma( stop ), stopNode->coords );
	if (allowed_signatures.size() >= MaxClassesToExplore  && homotopy_weight_in_cost > 0.0)
		for (a=0; a<Active_HomotopySearchProblem_p->S; a++)
			stopNode->homotopy_signature[a] = _round( allowed_signatures[0][a] * HomotopySignatureMultiplier );
	else
		homotopy_weight_in_cost = -1.0;
	
	// Allowed and blocked homotopy signatures
	for (a=0; a<allowed_signatures.size(); a++)
	{
		int* thisAllowed = new int[Active_HomotopySearchProblem_p->S];
		for (int b=0; b<Active_HomotopySearchProblem_p->S; b++)
			thisAllowed[b] = _round( allowed_signatures[a][b] * HomotopySignatureMultiplier );
		allowed_homotopy_signatures.push_back( thisAllowed );
	}
	
	for (a=0; a<blocked_signatures.size(); a++)
	{
		int* thisBlocked = new int[Active_HomotopySearchProblem_p->S];
		for (int b=0; b<Active_HomotopySearchProblem_p->S; b++)
			thisBlocked[b] = _round( blocked_signatures[a][b] * HomotopySignatureMultiplier );
		blocked_homotopy_signatures.push_back( thisBlocked );
	}
	
	// Create instance of GenericSearchGraphDescriptor
	graph = new GenericSearchGraphDescriptor<homotopyNode,int>;
	
	graph->hashTableSize = hashTableSize;
	graph->func_container = this;
	
	graph->SeedNodes.push_back(*startNode);
	graph->TargetNode = *stopNode;
	
	// Create instance of planner and plan
	planner = new A_star_planner<homotopyNode,int>;
	
	if (doReplan) printf("--- Planning the first round for pre-compuatation... \n");
	planner->setParams( A_star_suboptimality_eps, A_star_heap_key_count, A_star_progress_display_interval );
	planner->init(*graph);
	planner->plan();
	if (doReplan)
	{
		printf("--- Main planning... \n");
		storedPathsCount = 0;
		planner->clearLastPlanAndInit();
		planner->plan();
	}
	
	// Populate solution trajectories
	plannedTrajs.clear();
	plannedTrajSignatures.clear();
	std::vector< std::vector< homotopyNode > > paths = planner->getPlannedPaths();
	arma::vec temp_sig(Active_HomotopySearchProblem_p->S);
	for (a=0; a<paths.size(); a++)
	{
		std::vector< arma::vec > thisTraj;
		for (int b=paths[a].size()-1; b>=0; b--)
			thisTraj.push_back( dcTransform.disc2cont_arma( Array2StdVec<int>(paths[a][b].coords, Active_HomotopySearchProblem_p->D) ) );
		std::vector<int> ScaledSignature = Array2StdVec<int>(paths[a][0].homotopy_signature, Active_HomotopySearchProblem_p->S);
		for (int b=0; b<Active_HomotopySearchProblem_p->S; b++)
			temp_sig[b] = (double)ScaledSignature[b] / (double)HomotopySignatureMultiplier;
		plannedTrajSignatures.push_back( temp_sig );
		plannedTrajs.push_back( thisTraj );
	}
	
	delete graph;
	delete planner;
}

// --------

void gsl_err_handler(const char* reason, const char* file, int line, int gsl_errno)
{
	Active_HomotopySearchProblem_p->wasErrorOnLastIntegration = true;
	strcpy( Active_HomotopySearchProblem_p->errorReasonOnLastIntegration, reason );
}


// =========================================


