#ifndef __HOMOTOPY_NN_ASTAR_2F585H2B321R_H_
#define __HOMOTOPY_NN_ASTAR_2F585H2B321R_H_


// Standard includes
#include <stdio.h>
#include <math.h>
#include <vector>
#include <cstdlib>
#include <armadillo>
#include <ctime>
#include "homotopy_tools.h"
#include "topology_tools.h"
#include "subh_utils.h"
#include "yagsbpl_base.h"
#include "planners/A_star.h"


class homotopyNode;
class HomotopySearchProblem;

// ============================================================
// A homotopy problem needs to create and provide an instance of this class
// IMPORTANT: Only one instance of "HomotopySearchProblem" should be maintained in a program at a time.
//				This is because of a few ststic members in the class.

static HomotopySearchProblem* Active_HomotopySearchProblem_p;

class HomotopySearchProblem : public SearchGraphDescriptorFunctionContainer<homotopyNode,int>
{
public:
	// Search space and discretization descriptions
	DiscContTransform dcTransform;					// From "subh_utils.h"
	int HomotopySignatureMultiplier;	// optional. Default value set by constructor.
	int CostMultiplier;					// optional. Default value set by constructor.
	
	int HomotopySignatureCompareThresh;
	int GraphDiameter; // used to determine HomotopySignatureCompareThresh.
	
	// Homotopy class problem description
	HomotopyProblem* homotopyProb;					// From "homotopy_tools.h"
	
	// Used in Searching problems
	arma::vec start; 				// D-dimentional start coordinate
	arma::vec stop;					// D-dimentional stop coordinate
	
	// Function for checking if a coordinate is accessible
	bool (*isCoordFree)(arma::vec coord);
	
	// Actions - to be determined by one of the following:
	std::vector<arma::vec> forwardSteps;									// Set length 0 to disable
	std::vector<arma::vec> (*forwardSteps_f)(arma::vec coord);			// Set to NULL to disable
	
	// Cost - to be determined by one of the following:
	std::vector<double> stepCosts;								// Should be same size as "forwardSteps". Set length 0 to disable
	arma::mat (*metricTensor_f)(arma::vec coord);			// Set to NULL to disable
	double (*actionCost_f)(arma::vec coord, arma::vec action);
	
	// Heuristics
	double (*heuristics_f)(arma::vec c1, arma::vec c2);		// Set to NULL for 0 heuristics
	
	// Homotopy class constraints and/or exploration.
	int MaxClassesToExplore; 				// Set to 0 to disable exploration and enable planning with constraint instead
	std::vector<arma::vec> allowed_signatures;	// Set length to 0 to disable (i.e. set to complement of "blocked_signatures")
	std::vector<arma::vec> blocked_signatures;	// Set length to 0 to disable (i.e. set to complement of "allowed_signatures")
	
	// The following is for weighting edge costs with norm of the homotopy signature of the edge.
	//   Will be used only when MaxClassesToExplore == allowed_signatures.size().
	//   Assumes that allowed_signatures contain the paths in order of the net path lengths.
	double homotopy_weight_in_cost; // Set to negative to forcefully disable (default)
	
	// Other optional parameters
	bool onlyExploreJordan;						// In exploration problem, only explore Jordan paths. false by default
	bool discardNonjordanStates;
	double A_star_suboptimality_eps; 			// Do not set to keep it 1.0
	int A_star_heap_key_count, A_star_progress_display_interval;
	
	
	// ===========================================
	// ===========================================
	// Main members
	GenericSearchGraphDescriptor<homotopyNode,int>* graph;
	A_star_planner<homotopyNode,int>* planner;
	bool wasErrorOnLastIntegration;
	char errorReasonOnLastIntegration[1024];
	bool doReplan; // Plans twice - emulates precomputation of edges
	// -------------------------------------------
	// Helper functions and variables
	static unsigned int shuffle(unsigned int key);
	int hashTableSize;
	double HomotopyCostComponent(homotopyNode& n1, homotopyNode& n2);
	
	// -------------------------------------------
	// Functions derived from the user-specified quantities for use in GenericSearchGraphDescriptor<homotopyNode,int>"
	int getHashBin(homotopyNode& n);
	bool isAccessible(homotopyNode& n); // this also checks for base node in hash
	void getSuccessors(homotopyNode& n, std::vector<homotopyNode>* s, std::vector<int>* costs); // this also checks for base node in hash
	int getHeuristics(homotopyNode& n1, homotopyNode& n2);
	bool storePath(homotopyNode& n);
	bool stopSearch(homotopyNode& n);
	// Derived variables
	int storedPathsCount;
	homotopyNode* startNode;
	homotopyNode* stopNode;
	std::vector<int*> allowed_homotopy_signatures;
	std::vector<int*> blocked_homotopy_signatures;
	
	// -------------------------------------------
	// Constructor and other variables
	HomotopySearchProblem();
	void init_and_plan(void);	// To be called after all quantities are set. Returns trajectories.
	std::vector< std::vector< arma::vec > > plannedTrajs;
	std::vector< arma::vec > plannedTrajSignatures;
	
	// -------------------------------------------
	// Constant value members from the active instance to be used by coordNode, homotopyNode
	int D; // Dimension of ambient manifold
	int S; // Number of singularity manifolds
};


void gsl_err_handler(const char* reason, const char* file, int line, int gsl_errno);

/*
static int HomotopySearchProblem::HomotopySignatureCompareThresh;
static int HomotopySearchProblem::D;
static int HomotopySearchProblem::S; */

// ============================================================

/*
class coordNode
{
public:
	int* coords;
	bool operator==(const coordNode& n);
	coordNode() { coords = new int [HomotopySearchProblem::D]; };
	~coordNode() { delete[] coords; };
};

class coordNodeMeasure
{
public:
	int cc;
	int* hh;
	const coordNodeTransition operator+(const coordNodeTransition& nt) const;
	coordNodeMeasure() { hh = new int [HomotopySearchProblem::S]; };
	~coordNodeMeasure() { delete[] hh; };
}; */

class homotopyNode
{
public:
	int* coords;				// For efficiency, std::vector is not used
	int* homotopy_signature;
	
	bool operator==(const homotopyNode& n);
	bool isBaseCoordsSame(homotopyNode& n);
	friend homotopyNode operator+(const homotopyNode& lhs, const homotopyNode& rhs); // addition of coordinates and homotopy signature
	friend homotopyNode operator-(const homotopyNode& lhs, const homotopyNode& rhs); // substraction of coordinates and homotopy signature
	
	homotopyNode();
	homotopyNode(const homotopyNode& other);
	homotopyNode& operator=(const homotopyNode& other);
	~homotopyNode() { delete[] coords; delete[] homotopy_signature; };
	
	void print(void); // used for debugging
};

// ============================================================

/*
class homotopy_generic
{
public:
	HomotopySearchProblem searchProb;
	
	// "Base Graph Descriptor" - Container and functions for coordNode
	GenericSearchGraphDescriptor<coordNode,coordNodeMeasure>* BaseGraphDescriptor;
	HashTableContainer<coordNode,coordNodeMeasure,int>* hash;
	// Functions for BaseGraphDescriptor derived from searchProb
	int getHashBin_bg(coordNode& n);
	bool isAccessible_bg(coordNode& n);
	void getSuccessors_bg( coordNode& n, std::vector<coordNode>* s, std::vector<coordNodeMeasure>* c);
	
	homotopy_generic (HomotopySearchProblem sp )
	{
		searchProb = sp;
		
		// Create "BaseGraphDescriptor" from "searchProb"
		BaseGraphDescriptor = new GenericSearchGraphDescriptor<coordNode,coordNodeMeasure>;
		*BaseGraphDescriptor = ;
	
		hash = new HashTableContainer<NodeType,CostType,PlannerSpecificVariables>;
		hash->friendGraphDescriptor_p = GraphDescriptor;
		hash->init_HastTable( GraphDescriptor->hashTableSize );
	};
} */

#endif

